from amuse.test.amusetest import TestWithMPI
import os
import sys
import numpy
import math

from amuse.community.aarsethzare.interface import AarsethZareInterface, AarsethZare

from amuse.units import nbody_system
from amuse.units import units
from amuse import datamodel
from amuse.ic import plummer
try:
    from matplotlib import pyplot
    HAS_MATPLOTLIB = True
except ImportError:
    HAS_MATPLOTLIB = False


class TestAarsethZareInterface(TestWithMPI):

    def test0(self):
        instance = AarsethZareInterface()
        instance.stop()

    def test1(self):

        instance = AarsethZareInterface()
        time, x, y, z, vx, vy, vz, error = instance.evolve_triple(
            [0, 0, 0],
            [1, 1, 1],
            [-10, 0, 10],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [1.0, 1.0, 1.0]
        )

        print(time, x, y, z, vx, vy, vz, error)
        self.assertEqual(error, 0)
        self.assertAlmostRelativeEquals(y, [0, 0, 0])
        self.assertAlmostRelativeEquals(z, [0, 0, 0])
        self.assertAlmostRelativeEquals(x[-1], -x[0], 10)
        self.assertAlmostRelativeEquals(x[1], 0, 10)
        self.assertAlmostRelativeEquals(vx[-1], -vx[0], 10)
        self.assertAlmostRelativeEquals(vx[1], 0, 10)
        self.assertAlmostRelativeEquals(vy, [0, 0, 0])
        self.assertAlmostRelativeEquals(vz, [0, 0, 0])

    def test2(self):

        instance = AarsethZareInterface()
        time, x, y, z, vx, vy, vz, error = instance.evolve_triple(
            [0, 0],
            [1, 1],
            [-10, 0],
            [0, 0],
            [0, 0],
            [0, 0],
            [0, 0],
            [0, 0],
            [1.0]
        )

        self.assertEqual(error, -1)


class TestAarsethZare(TestWithMPI):
    def new_system_of_sun_and_earth_and_moon(self):
        stars = datamodel.Stars(3)
        sun = stars[0]
        sun.mass = units.MSun(1.0)
        sun.position = units.m(numpy.array((0.0, 0.0, 0.0)))
        sun.velocity = units.ms(numpy.array((0.0, 0.0, 0.0)))
        sun.radius = units.RSun(1.0)

        earth = stars[1]
        earth.mass = units.kg(5.9736e24)
        earth.radius = units.km(6371)
        earth.position = units.km(numpy.array((149.5e6, 0.0, 0.0)))
        earth.velocity = units.ms(numpy.array((0.0, 29800, 0.0)))

        moon = stars[2]
        moon.mass = units.kg(7.3477e22)
        moon.radius = units.km(1737.10)
        moon.position = units.km(numpy.array((149.5e6 + 384399.0, 0.0, 0.0)))
        moon.velocity = ([0.0, 1.022, 0] | units.km/units.s) + earth.velocity
        return stars

    def test0(self):
        instance = AarsethZare()
        instance.stop()

    def test1(self):
        instance = AarsethZare()
        time, x, y, z, vx, vy, vz = instance.evolve_triple(
            [0, 0, 0] | nbody_system.time,
            [1, 1, 1] | nbody_system.mass,
            [-10, 0, 10] | nbody_system.length,
            [0, 0, 0] | nbody_system.length,
            [0, 0, 0] | nbody_system.length,
            [0, 0, 0] | nbody_system.speed,
            [0, 0, 0] | nbody_system.speed,
            [0, 0, 0] | nbody_system.speed,
            [1.0, 1.0, 1.0] | nbody_system.time
        )

        self.assertAlmostRelativeEquals(y, [0, 0, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(z, [0, 0, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(x[-1], -x[0], 10)
        self.assertAlmostRelativeEquals(x[1], 0 | nbody_system.length, 10)
        self.assertAlmostRelativeEquals(vx[-1], -vx[0], 10)
        self.assertAlmostRelativeEquals(vx[1], 0 | nbody_system.speed, 10)
        self.assertAlmostRelativeEquals(vy, [0, 0, 0] | nbody_system.speed)
        self.assertAlmostRelativeEquals(vz, [0, 0, 0] | nbody_system.speed)

    def test2(self):
        instance = AarsethZare()

        particles = datamodel.Particles(3)
        particles.mass = 1.0 | nbody_system.mass
        particles.position = [
                [-10.0, 0.0, 0.0],
                [0.0, 0.0, 0.0],
                [10.0, 0.0, 0.0],
            ] | nbody_system.length
        particles.velocity = [0.0, 0.0, 0.0] | nbody_system.speed

        instance.particles.add_particles(particles)
        instance.evolve_model(1.0 | nbody_system.time)

        self.assertAlmostRelativeEquals(instance.particles.y, [0, 0, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(instance.particles.z, [0, 0, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(instance.particles.x[-1], -instance.particles.x[0], 10)
        self.assertAlmostRelativeEquals(instance.particles.x[1], 0 | nbody_system.length, 10)
        self.assertAlmostRelativeEquals(instance.particles.vx[-1], -instance.particles.vx[0], 10)
        self.assertAlmostRelativeEquals(instance.particles.vx[1], 0 | nbody_system.speed, 10)
        self.assertAlmostRelativeEquals(instance.particles.vy, [0, 0, 0] | nbody_system.speed)
        self.assertAlmostRelativeEquals(instance.particles.vz, [0, 0, 0] | nbody_system.speed)

    def test3(self):
        instance = AarsethZare(nbody_system.nbody_to_si(1.0 | units.yr, 1.0 | units.MSun))

        particles = self.new_system_of_sun_and_earth_and_moon()
        instance.particles.add_particles(particles)

        earth = instance.particles[1]

        position_at_start = earth.position[0]

        instance.evolve_model(365.0 | units.day)

        position_after_full_rotation = earth.position[0]

        self.assertAlmostRelativeEquals(position_at_start, position_after_full_rotation, 4)

        instance.evolve_model(1.5 | units.yr)

        position_after_half_a_rotation = earth.position[0]

        self.assertAlmostRelativeEquals(-position_at_start, position_after_half_a_rotation, 2)

        instance.evolve_model(1.75 | units.yr)

        position_after_quarter_a_rotation = earth.position[1]

        self.assertAlmostRelativeEquals(position_at_start, position_after_quarter_a_rotation, 2)

        instance.stop()



import os
import os.path
import math

from amuse.community import *
from amuse.test.amusetest import TestWithMPI
from amuse.units import units, nbody_system, constants
from amuse.datamodel import Particles

from amuse.community.adaptb.interface import AdaptbInterface, Adaptb
from amuse.community.adaptb.functions import read_log, read_out, read_xy


class TestAdaptbInterface(TestWithMPI):

    def test1(self):
        print("Test AdaptbInterface initialization")
        instance = self.new_instance_of_an_optional_code(AdaptbInterface)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_adaptb_output_directory(instance.output_directory))
        self.assertEqual(0, instance.commit_parameters())
        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test2(self):
        print("Test AdaptbInterface new_particle / get_state")
        instance = self.new_instance_of_an_optional_code(AdaptbInterface)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_adaptb_output_directory(instance.output_directory))
        self.assertEqual(0, instance.commit_parameters())

        id, error = instance.new_particle(mass=11.0, radius=2.0, x=0.0, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0)
        self.assertEqual(0, error)
        self.assertEqual(0, id)
        id, error = instance.new_particle(mass=21.0, radius=5.0, x=10.0, y=0.0, z=0.0, vx=10.0, vy=0.0, vz=0.0)
        self.assertEqual(0, error)
        self.assertEqual(1, id)
        self.assertEqual(0, instance.commit_particles())

        retrieved_state1 = instance.get_state(0)
        retrieved_state2 = instance.get_state(1)
        self.assertEqual(0,  retrieved_state1['__result'])
        self.assertEqual(0,  retrieved_state2['__result'])
        self.assertEqual(11.0,  retrieved_state1['mass'])
        self.assertEqual(21.0,  retrieved_state2['mass'])
        self.assertEqual(0.0,  retrieved_state1['x'])
        self.assertEqual(10.0,  retrieved_state2['x'])

        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test4(self):
        print("Test AdaptbInterface particle property getters/setters")
        instance = self.new_instance_of_an_optional_code(AdaptbInterface)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_adaptb_output_directory(instance.output_directory))
        self.assertEqual(0, instance.commit_parameters())
        self.assertEqual([0, 0], list(instance.new_particle(0.01,  1, 0, 0,  0, 1, 0, 0.1).values()))
        self.assertEqual([1, 0], list(instance.new_particle(0.02, -1, 0, 0,  0, -1, 0, 0.1).values()))
        self.assertEqual(0, instance.commit_particles())

        # getters
        mass, result = instance.get_mass(0)
        self.assertAlmostEqual(0.01, mass)
        self.assertEqual(0, result)
        radius, result = instance.get_radius(1)
        self.assertAlmostEqual(0.1, radius)
        self.assertEqual(0, result)
        self.assertEqual(-3, instance.get_mass(2)['__result'])  # Particle not found
        self.assertEqual([1, 0, 0,  0], list(instance.get_position(0).values()))
        self.assertEqual([-1, 0, 0,  0], list(instance.get_position(1).values()))
        self.assertEqual([0, 1, 0,  0], list(instance.get_velocity(0).values()))
        self.assertEqual([0, -1, 0,  0], list(instance.get_velocity(1).values()))

        # setters
        self.assertEqual(0, instance.set_state(0, 0.01, 1, 2, 3, 4, 5, 6, 0.1))
        self.assertEqual([0.01, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 0.1, 0], list(instance.get_state(0).values()))
        self.assertEqual(0, instance.set_mass(0, 0.02))
        self.assertEqual([0.02, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 0.1, 0], list(instance.get_state(0).values()))
        self.assertEqual(0, instance.set_radius(0, 0.2))
        self.assertEqual([0.02, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 0.2, 0], list(instance.get_state(0).values()))
        self.assertEqual(0, instance.set_position(0, 10, 20, 30))
        self.assertEqual([0.02, 10.0, 20.0, 30.0, 4.0, 5.0, 6.0, 0.2, 0], list(instance.get_state(0).values()))
        self.assertEqual(0, instance.set_velocity(0, 40, 50, 60))
        self.assertEqual([0.02, 10.0, 20.0, 30.0, 40.0, 50.0, 60.0, 0.2, 0], list(instance.get_state(0).values()))

        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test5(self):
        print("Test AdaptbInterface parameters")
        instance = self.new_instance_of_an_optional_code(AdaptbInterface)
        self.assertEqual(0, instance.initialize_code())

        # word length
        self.assertEqual([64, 0], list(instance.get_word_length().values()))
        self.assertEqual(0, instance.set_word_length(80))
        self.assertEqual([80, 0], list(instance.get_word_length().values()))

        # bs tolerance, default (double) implementation
        self.assertEqual([1.0e-6, 0], list(instance.get_bs_tolerance_float64().values()))
        self.assertEqual(0, instance.set_bs_tolerance_float64(1.0e-8))
        self.assertEqual([1.0e-8, 0], list(instance.get_bs_tolerance_float64().values()))
        # bs tolerance, string implementation for values requiring higher precision (note: actual accuracy depends on word_length)
        self.assertEqual(1e-8, eval(instance.get_bs_tolerance_string()["bs_tolerance"]))
        self.assertEqual(0, instance.set_bs_tolerance_string("1e-10"))
        self.assertEqual([1e-10, 0], list(instance.get_bs_tolerance_float64().values()))
        # new mpreal 1e-10 prints as 9.99999...e-11
        # self.assertEquals(["1e-10", 0], instance.get_bs_tolerance_string().values())

        # softening
        self.assertEqual([0.0, 0], list(instance.get_eps2().values()))
        self.assertEqual(0, instance.set_eps2(2e-1))
        self.assertEqual([0.2, 0], list(instance.get_eps2().values()))

        # print intervals
        self.assertEqual([1e-1, 0], list(instance.get_dt_print().values()))
        self.assertEqual(0, instance.set_dt_print(1e-2))
        self.assertEqual([1e-2, 0], list(instance.get_dt_print().values()))

        # max cpu time
        self.assertEqual([3600, 0], list(instance.get_max_cpu_time().values()))
        self.assertEqual(0, instance.set_max_cpu_time(120))
        self.assertEqual([120, 0], list(instance.get_max_cpu_time().values()))

        # output dir
        self.assertEqual(["./", 0], list(instance.get_adaptb_output_directory().values()))
        self.assertEqual(0, instance.set_adaptb_output_directory("./out"))
        self.assertEqual(["./out/", 0], list(instance.get_adaptb_output_directory().values()))
        self.assertEqual(0, instance.set_adaptb_output_directory(instance.output_directory))
        self.assertEqual([instance.output_directory+"/", 0], list(instance.get_adaptb_output_directory().values()))

        self.assertEqual(0, instance.commit_parameters())
        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test6(self):
        print("Test AdaptbInterface evolve_model, equal-mass binary")
        instance = self.new_instance_of_an_optional_code(AdaptbInterface)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_dt_print(1e-1))
        self.assertEqual(0, instance.set_word_length(64))
        self.assertEqual(0, instance.set_bs_tolerance_float64(1.0e-8))
        self.assertEqual(0, instance.set_adaptb_output_directory(instance.output_directory))
        self.assertEqual(0, instance.commit_parameters())

        self.assertEqual([0, 0], list(instance.new_particle(0.5,  0.5, 0, 0,  0, 0.5, 0).values()))
        self.assertEqual([1, 0], list(instance.new_particle(0.5, -0.5, 0, 0,  0, -0.5, 0).values()))
        self.assertEqual(0, instance.commit_particles())

        self.assertEqual(0, instance.evolve_model(math.pi))  # half an orbit
        for result, expected in zip(list(instance.get_position(0).values()), [-0.5, 0.0, 0.0, 0]):
            self.assertAlmostEqual(result, expected, 5)

        self.assertEqual(0, instance.evolve_model(2 * math.pi))  # full orbit
        for result, expected in zip(list(instance.get_position(0).values()), [0.5, 0.0, 0.0, 0]):
            self.assertAlmostEqual(result, expected, 5)

        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test7(self):
        print("Test AdaptbInterface evolve_model, pythagorean problem")
        instance = self.new_instance_of_an_optional_code(AdaptbInterface)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_dt_print(10.0))
        self.assertEqual(0, instance.set_bs_tolerance_float64(1.0e-2))
        self.assertEqual(0, instance.set_word_length(64))
        self.assertEqual(0, instance.set_adaptb_output_directory(instance.output_directory))
        self.assertEqual(0, instance.commit_parameters())

        self.assertEqual([0, 0], list(instance.new_particle("3",  "1",  "3", "0", "0", "0", "0").values()))
        self.assertEqual([1, 0], list(instance.new_particle("4", "-2", "-1", "0", "0", "0", "0").values()))
        self.assertEqual([2, 0], list(instance.new_particle("5",  "1", "-1", "0", "0", "0", "0").values()))
        self.assertEqual(0, instance.commit_particles())
        self.assertEqual(0, instance.evolve_model(100))

        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def slowtest8(self):
        print("Test AdaptbInterface evolve_model, pythagorean problem, show convergence")
        tolerance = [1e-0, 1e-2, 1e-4, 1e-6, 1e-8, 1e-10, 1e-12, 1e-14, 1e-16, 1e-18, 1e-20, 1e-22, 1e-24]
        word_length = [64,    64,     64,     64,     64,     64,      80,      80,      96,      96,      112,    112,     128]

        for tol, word_len in zip(tolerance, word_length):
            instance = self.new_instance_of_an_optional_code(AdaptbInterface)
            self.assertEqual(0, instance.initialize_code())
            self.assertEqual(0, instance.set_dt_print(0.1))
            self.assertEqual(0, instance.set_bs_tolerance_float64(tol))
            self.assertEqual(0, instance.set_word_length(word_len))
            self.assertEqual(0, instance.set_adaptb_output_directory(instance.output_directory))
            self.assertEqual(0, instance.commit_parameters())
            self.assertEqual([0, 0], list(instance.new_particle("3", "0",  "1",  "3", "0", "0", "0", "0").values()))
            self.assertEqual([1, 0], list(instance.new_particle("4", "0", "-2", "-1", "0", "0", "0", "0").values()))
            self.assertEqual([2, 0], list(instance.new_particle("5", "0",  "1", "-1", "0", "0", "0", "0").values()))
            self.assertEqual(0, instance.commit_particles())
            self.assertEqual(0, instance.evolve_model(100))
            self.assertEqual(0, instance.cleanup_code())
            instance.stop()

            data_log = read_log(os.path.join(instance.output_directory, "file.log"))
            data_out = read_out(os.path.join(instance.output_directory, "file.out"))
            tcpu = data_log[9]
            dE = data_log[10]
            x0 = data_out[0]
            vx0 = data_out[1]
            print("\nTolerance:", tol)
            print("Word length:", word_len)
            print("CPU time for calculation:", tcpu)
            print("Relative energy error:", dE)
            print("Final x position and velocity of first particle:", x0, vx0)

    def xtest9(self):
        print("Test AdaptbInterface evolve_model, pythagorean problem, plot orbits")
        tolerance = ["1e-4", "1e-8"]  # , "1e-12", "1e-16"]
        word_length = [64,     64]  # ,      80,      96]

        x1 = []
        y1 = []
        x2 = []
        y2 = []
        x3 = []
        y3 = []

        i = 0
        while i < len(tolerance):
            instance = self.new_instance_of_an_optional_code(AdaptbInterface)
            self.assertEqual(0, instance.initialize_code())
            self.assertEqual(0, instance.set_dt_print(1e-1))
            self.assertEqual(0, instance.set_bs_tolerance(tolerance[i]))
            self.assertEqual(0, instance.set_word_length(word_length[i]))
            self.assertEqual(0, instance.set_adaptb_output_directory(instance.output_directory))
            self.assertEqual(0, instance.commit_parameters())

            self.assertEqual([0, 0], list(instance.new_particle("3", "0",  "1",  "3", "0", "0", "0", "0").values()))
            self.assertEqual([1, 0], list(instance.new_particle("4", "0", "-2", "-1", "0", "0", "0", "0").values()))
            self.assertEqual([2, 0], list(instance.new_particle("5", "0",  "1", "-1", "0", "0", "0", "0").values()))
            self.assertEqual(0, instance.commit_particles())

            self.assertEqual(0, instance.evolve_model(100))

            self.assertEqual(0, instance.cleanup_code())
            instance.stop()

            file_out = "file.out"
            my_x1, my_y1, my_x2, my_y2, my_x3, my_y3 = read_xy(file_out)

            x1.append(my_x1)
            y1.append(my_y1)
            x2.append(my_x2)
            y2.append(my_y2)
            x3.append(my_x3)
            y3.append(my_y3)

            i += 1

        print(x1)


class TestAdaptb(TestWithMPI):

    def new_sun_earth_system(self):
        particles = Particles(2)
        particles.mass = [1.0, 3.0037e-6] | units.MSun
        particles.radius = 1.0 | units.RSun
        particles.position = [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]] | units.AU
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | units.km / units.s
        particles[1].vy = (constants.G * particles.total_mass() / (1.0 | units.AU)).sqrt()
        return particles

    def test1(self):
        print("Testing Adaptb initialization")
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)
        instance = self.new_instance_of_an_optional_code(Adaptb, convert_nbody)
        instance.initialize_code()
        instance.commit_parameters()
        instance.cleanup_code()
        instance.stop()

    def test2(self):
        print("Testing Adaptb parameters")
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)
        instance = self.new_instance_of_an_optional_code(Adaptb, convert_nbody)
        instance.initialize_code()

# ~        print instance.parameters
        self.assertEqual(instance.parameters.bs_tolerance, 1.0e-6)
        instance.parameters.bs_tolerance = 1.0e-9
        self.assertEqual(instance.parameters.bs_tolerance, 1.0e-9)

        self.assertEqual(instance.parameters.epsilon_squared, 0.0 | units.m**2)
        instance.parameters.epsilon_squared = 1.0e-4 | nbody_system.length**2
        self.assertEqual(instance.parameters.epsilon_squared, convert_nbody.to_si(1.0e-4 | nbody_system.length**2))

        self.assertEqual(instance.parameters.word_length, 64)
        instance.parameters.word_length = 128
        self.assertEqual(instance.parameters.word_length, 128)

        self.assertEqual(instance.parameters.dt_print, convert_nbody.to_si(0.1 | nbody_system.time))
        instance.parameters.dt_print = 1.0e-4 | nbody_system.time
        self.assertEqual(instance.parameters.dt_print, convert_nbody.to_si(1.0e-4 | nbody_system.time))

        self.assertEqual(instance.parameters.adaptb_output_directory, instance.output_directory + os.sep)
        instance.parameters.adaptb_output_directory = "./out"
        self.assertEqual(instance.parameters.adaptb_output_directory, "./out/")
        instance.parameters.adaptb_output_directory = instance.output_directory
        self.assertEqual(instance.parameters.adaptb_output_directory, instance.output_directory + os.sep)

        self.assertEqual(instance.parameters.time_limit_cpu, 3600.0 | units.s)
        instance.parameters.time_limit_cpu = 7200.0 | units.s
        self.assertEqual(instance.parameters.time_limit_cpu, 7200.0 | units.s)

        instance.stop()

    def test3(self):
        print("Testing Adaptb particles")
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)
        instance = self.new_instance_of_an_optional_code(Adaptb, convert_nbody)
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(self.new_sun_earth_system())
        instance.commit_particles()

        self.assertAlmostEqual(instance.particles.mass, [1.0, 3.0037e-6] | units.MSun)
        self.assertAlmostEqual(instance.particles.radius, 1.0 | units.RSun)
        self.assertAlmostEqual(instance.particles.position,
            [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]] | units.AU)
        self.assertAlmostEqual(instance.particles.velocity,
            [[0.0, 0.0, 0.0], [0.0, 29.7885, 0.0]] | units.km / units.s, 3)

        instance.cleanup_code()
        instance.stop()

    def test4(self):
        print("Testing Adaptb evolve_model, 2 particles")
        particles = Particles(2)
        particles.mass = 0.5 | units.MSun
        particles.radius = 1.0 | units.RSun
        particles.position = [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]] | units.AU
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | units.km / units.s
        particles[1].vy = (constants.G * (1.0 | units.MSun) / (1.0 | units.AU)).sqrt()
        particles.move_to_center()

        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)
        instance = self.new_instance_of_an_optional_code(Adaptb, convert_nbody)
        instance.initialize_code()
        instance.parameters.dt_print = 0.1 | units.yr
        instance.parameters.bs_tolerance = 1.0e-8
        instance.commit_parameters()
        instance.particles.add_particles(particles)
        instance.commit_particles()
        primary = instance.particles[0]

        P = 2 * math.pi * primary.x / primary.vy

        position_at_start = primary.position.x
        instance.evolve_model(P / 4.0)
        self.assertAlmostRelativeEqual(position_at_start, primary.position.y, 6)

        instance.evolve_model(P / 2.0)
        self.assertAlmostRelativeEqual(position_at_start, -primary.position.x, 6)

        instance.evolve_model(P)
        self.assertAlmostRelativeEqual(position_at_start, primary.position.x, 6)

        instance.cleanup_code()
        instance.stop()



import numpy

from amuse.test import amusetest

from amuse.ext.orbital_elements import (
    generate_binaries,
    get_orbital_elements_from_binaries,
    new_binary_from_orbital_elements,
    # get_orbital_elements_from_binary,
    orbital_elements_for_rel_posvel_arrays,
    orbital_elements,
    rel_posvel_arrays_from_orbital_elements,
)

from amuse.units import units
from amuse.units import constants
from amuse.units import nbody_system
from amuse import datamodel

from numpy import random


class KeplerTests(amusetest.TestCase):

    def test1(self):
        mass1 = 1 | nbody_system.mass
        mass2 = 1 | nbody_system.mass

        binary = new_binary_from_orbital_elements(
            mass1,
            mass2,
            1 | nbody_system.length
        )

        self.assertEqual(len(binary), 2)

        binary.position -= binary[0].position
        binary.velocity -= binary[0].velocity

        self.assertAlmostRelativeEquals(
                binary[0].position,
                [0, 0, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(
                binary[1].position,
                [1, 0, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(
                binary[0].velocity,
                [0, 0, 0] | nbody_system.speed)
        self.assertAlmostRelativeEquals(
                binary[1].velocity,
                [0, numpy.sqrt(2), 0] | nbody_system.speed)

    def test2(self):
        # test going around in a circular orbit
        mass1 = 1 | nbody_system.mass
        mass2 = 1 | nbody_system.mass

        binary = new_binary_from_orbital_elements(
            mass1,
            mass2,
            1 | nbody_system.length,
            eccentricity=0,
            true_anomaly=90,
        )

        self.assertEqual(len(binary), 2)

        binary.position -= binary[0].position
        binary.velocity -= binary[0].velocity

        self.assertAlmostRelativeEquals(
                binary[0].position,
                [0, 0, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(
                binary[1].position,
                [0, 1, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(
                binary[0].velocity,
                [0, 0, 0] | nbody_system.speed)
        self.assertAlmostRelativeEquals(
                binary[1].velocity,
                [-numpy.sqrt(2), 0, 0] | nbody_system.speed)

        binary = new_binary_from_orbital_elements(
            mass1,
            mass2,
            1 | nbody_system.length,
            eccentricity=0,
            true_anomaly=180,
        )

        self.assertEqual(len(binary), 2)

        binary.position -= binary[0].position
        binary.velocity -= binary[0].velocity

        self.assertAlmostRelativeEquals(
                binary[0].position, [0, 0, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(
                binary[1].position, [-1, 0, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(
                binary[0].velocity, [0, 0, 0] | nbody_system.speed)
        self.assertAlmostRelativeEquals(
                binary[1].velocity, [0, -numpy.sqrt(2), 0] | nbody_system.speed)

        binary = new_binary_from_orbital_elements(
            mass1,
            mass2,
            1 | nbody_system.length,
            eccentricity=0,
            true_anomaly=270,
        )

        self.assertEqual(len(binary), 2)

        binary.position -= binary[0].position
        binary.velocity -= binary[0].velocity

        self.assertAlmostRelativeEquals(binary[0].position, [0, 0, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(binary[1].position, [0, -1, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(binary[0].velocity, [0, 0, 0] | nbody_system.speed)
        self.assertAlmostRelativeEquals(binary[1].velocity, [numpy.sqrt(2), 0, 0] | nbody_system.speed)

        binary = new_binary_from_orbital_elements(
            mass1,
            mass2,
            1 | nbody_system.length,
            eccentricity=0,
            true_anomaly=45,
        )

        self.assertEqual(len(binary), 2)

        binary.position -= binary[0].position
        binary.velocity -= binary[0].velocity

        self.assertAlmostRelativeEquals(binary[0].position, [0, 0, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(binary[1].position, [0.5 * numpy.sqrt(2), 0.5 * numpy.sqrt(2), 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(binary[0].velocity, [0, 0, 0] | nbody_system.speed)
        self.assertAlmostRelativeEquals(binary[1].velocity, [-1, 1, 0] | nbody_system.speed)

    def test3(self):
        mass1 = 1. | nbody_system.mass
        mass2 = 1. | nbody_system.mass

        binary = new_binary_from_orbital_elements(
            mass1,
            mass2,
            1. | nbody_system.length,
            eccentricity=0.
        )

        self.assertEqual(len(binary), 2)
        self.assertAlmostRelativeEquals(binary[0].position, [-0.5, 0, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(binary[1].position, [0.5, 0, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(binary[0].velocity, [0, -1/numpy.sqrt(2), 0] | nbody_system.speed)
        self.assertAlmostRelativeEquals(binary[1].velocity, [0, 1/numpy.sqrt(2), 0] | nbody_system.speed)

    def test4(self):
        numpy.random.seed(3456789)
        N = 100

        mass1 = random.random(N) | nbody_system.mass
        mass2 = random.random(N) | nbody_system.mass
        semi_major_axis = (-numpy.log(random.random(N))) | nbody_system.length
        eccentricity = random.random(N)
        true_anomaly = (360.*random.random(N)-180.) | units.deg
        inclination = (180*random.random(N)) | units.deg
        longitude_of_the_ascending_node = (
                360*random.random(N)-180) | units.deg
        argument_of_periapsis = (360*random.random(N)-180) | units.deg

        for arg in zip(
                mass1, mass2, semi_major_axis, eccentricity, true_anomaly,
                inclination, longitude_of_the_ascending_node,
                argument_of_periapsis):
            arg_ = orbital_elements(
                    new_binary_from_orbital_elements(*arg))
            for i, (copy, org) in enumerate(zip(arg_, arg)):
                self.assertAlmostEqual(copy, org)

    def test5(self):
        numpy.random.seed(4567893)
        N = 100

        mass1 = random.random(N) | units.MSun
        mass2 = random.random(N) | units.MSun
        semi_major_axis = (-numpy.log(random.random(N))) | units.AU
        eccentricity = random.random(N)
        true_anomaly = (360.*random.random(N)-180.) | units.deg
        inclination = (180*random.random(N)) | units.deg
        longitude_of_the_ascending_node = (
                360*random.random(N)-180
                ) | units.deg
        argument_of_periapsis = (360*random.random(N)-180) | units.deg

        for arg in zip(
                mass1, mass2, semi_major_axis, eccentricity, true_anomaly,
                inclination, longitude_of_the_ascending_node,
                argument_of_periapsis):
            arg_ = orbital_elements(
                new_binary_from_orbital_elements(*arg, G=constants.G),
                G=constants.G)
            for i, (copy, org) in enumerate(zip(arg_, arg)):
                self.assertAlmostEqual(copy, org)

    def test_orbital_elements_of_binary_as_dict(self):
        numpy.random.seed(4567893)
        N = 100

        mass1 = random.random(N) | units.MSun
        mass2 = random.random(N) | units.MSun
        semi_major_axis = (-numpy.log(random.random(N))) | units.AU
        eccentricity = random.random(N)
        true_anomaly = (360.*random.random(N)-180.) | units.deg
        inclination = (180*random.random(N)) | units.deg
        longitude_of_the_ascending_node = (
                360*random.random(N)-180
                ) | units.deg
        argument_of_periapsis = (360*random.random(N)-180) | units.deg

        for arg in zip(
            mass1, mass2, semi_major_axis, eccentricity, true_anomaly,
            inclination, longitude_of_the_ascending_node, argument_of_periapsis
        ):
            arg_ = orbital_elements(
                new_binary_from_orbital_elements(*arg, G=constants.G),
                G=constants.G, return_dict=True)
            self.assertAlmostEqual(arg[0], arg_['mass_primary'])
            self.assertAlmostEqual(arg[1], arg_['mass_secondary'])
            self.assertAlmostEqual(arg[2], arg_['semi_major_axis'])
            self.assertAlmostEqual(arg[3], arg_['eccentricity'])
            self.assertAlmostEqual(arg[4], arg_['true_anomaly'])
            self.assertAlmostEqual(arg[5], arg_['inclination'])
            self.assertAlmostEqual(arg[6], arg_['longitude_of_the_ascending_node'])
            self.assertAlmostEqual(arg[7], arg_['argument_of_periapsis'])


    def test6(self):
        """
        testing orbital_elements_for_rel_posvel_arrays for N particles 
        with random orbital elements
        """
        numpy.random.seed(666)
        N = 100

        mass_sun = 1. | units.MSun
        mass1 = numpy.ones(N) * mass_sun
        mass2 = numpy.zeros(N) | units.MSun
        semi_major_axis = (-numpy.log(random.random(N))) | units.AU
        eccentricity = random.random(N)
        true_anomaly = 360.*random.random(N)-180.
        inclination = 180*random.random(N)
        longitude_of_the_ascending_node = 360*random.random(N)-180
        argument_of_periapsis = 360*random.random(N)-180

        comets = datamodel.Particles(N)
        for i, arg in enumerate(zip(mass1, mass2, semi_major_axis, eccentricity, true_anomaly, inclination,
                                   longitude_of_the_ascending_node, argument_of_periapsis)):
            sun_and_comet = new_binary_from_orbital_elements(*arg, G=constants.G)
            comets[i].mass = sun_and_comet[1].mass
            comets[i].position = sun_and_comet[1].position
            comets[i].velocity = sun_and_comet[1].velocity

        semi_major_axis_ext, eccentricity_ext, ta_ext, inclination_ext, \
        longitude_of_the_ascending_node_ext, argument_of_periapsis_ext = \
        orbital_elements_for_rel_posvel_arrays(comets.position,
                                               comets.velocity,
                                               comets.mass + mass_sun,
                                               G=constants.G)

        self.assertAlmostEqual(semi_major_axis, semi_major_axis_ext)
        self.assertAlmostEqual(eccentricity, eccentricity_ext)
        self.assertAlmostEqual(inclination, inclination_ext)
        self.assertAlmostEqual(longitude_of_the_ascending_node, longitude_of_the_ascending_node_ext)
        self.assertAlmostEqual(argument_of_periapsis, argument_of_periapsis_ext)
        self.assertAlmostEqual(true_anomaly, ta_ext)

    def test7(self):
        """
        testing orbital_elements_for_rel_posvel_arrays for the case of one particle
        """
        numpy.random.seed(999)

        mass1 = 0.5 | units.MSun
        mass2 = 0.8 | units.MSun
        sem = 12. | units.AU
        ecc = 0.05
        inc = 20.
        lon = 10.
        arg = 0.4
        ta = 360.*random.random()-180.

        binary = new_binary_from_orbital_elements(mass1,
                                                  mass2,
                                                  sem,
                                                  ecc,
                                                  ta,
                                                  inc,
                                                  lon,
                                                  arg,
                                                  G=constants.G)

        rel_pos = binary[1].position - binary[0].position
        rel_vel = binary[1].velocity - binary[0].velocity
        mass_12 = binary[1].mass + binary[0].mass
        sem_ext, ecc_ext, ta_ext, inc_ext, lon_ext, arg_ext = \
        orbital_elements_for_rel_posvel_arrays(rel_pos, rel_vel, mass_12, G=constants.G)

        self.assertAlmostEqual(sem, sem_ext)
        self.assertAlmostEqual(ecc, ecc_ext)
        self.assertAlmostEqual(inc, inc_ext)
        self.assertAlmostEqual(lon, lon_ext)
        self.assertAlmostEqual(arg, arg_ext)
        self.assertAlmostEqual(ta, ta_ext)

    def test8(self):
        """
        testing orbital_elements_for_rel_posvel_arrays for extreme cases
        """
        N = 3
        mass1 = (1.2*numpy.ones(N)) | units.MSun
        mass2 = (0.1, 0.05, 0.003) | units.MSun
        semi_major_axis = (1., 2., 3.) | units.AU
        eccentricity = (0., 0.5, 0.6)
        true_anomaly = (0., 0., 66.)
        inclination = (12., 0., 180.)
        longitude_of_the_ascending_node = (0., 0., 0.,)
        argument_of_periapsis = (0., 23., 90.)
        mass12 = []
        rel_position = []
        rel_velocity = []
        for i, arg in enumerate(zip(mass1, mass2, semi_major_axis, eccentricity, true_anomaly, inclination,
                                   longitude_of_the_ascending_node, argument_of_periapsis)):
            sun_and_comet = new_binary_from_orbital_elements(*arg, G=constants.G)
            mass12.append(sun_and_comet[0].mass + sun_and_comet[1].mass)
            rel_position.append(sun_and_comet[1].position - sun_and_comet[0].position)
            rel_velocity.append(sun_and_comet[1].velocity - sun_and_comet[0].velocity)

        # to convert lists to vector quantities
        rel_pos = numpy.array([vec_i.value_in(units.AU) for vec_i in rel_position]) | units.AU
        rel_vel = numpy.array([vec_i.value_in(units.kms) for vec_i in rel_velocity]) | units.kms
        mass_12 = numpy.array([m_i.value_in(units.MSun) for m_i in mass12]) | units.MSun

        semi_major_axis_ext, eccentricity_ext, ta_ext, inclination_ext, \
        longitude_of_the_ascending_node_ext, argument_of_periapsis_ext = \
        orbital_elements_for_rel_posvel_arrays(rel_pos,
                                               rel_vel,
                                               mass_12,
                                               G=constants.G)

        self.assertAlmostEqual(semi_major_axis, semi_major_axis_ext)
        self.assertAlmostEqual(eccentricity, eccentricity_ext)
        self.assertAlmostEqual(inclination, inclination_ext)
        self.assertAlmostEqual(longitude_of_the_ascending_node, longitude_of_the_ascending_node_ext)
        self.assertAlmostEqual(argument_of_periapsis, argument_of_periapsis_ext)
        self.assertAlmostEqual(true_anomaly, ta_ext)

    def test9(self):
        """
        testing orbital_elements_for_rel_posvel_arrays for N particles 
        with random orbital elements, nbody_system
        """
        numpy.random.seed(666)
        N = 100

        mass_sun = 1. | nbody_system.mass
        mass1 = numpy.ones(N) * mass_sun
        mass2 = numpy.zeros(N) | nbody_system.mass
        semi_major_axis = (-numpy.log(random.random(N))) | nbody_system.length
        eccentricity = random.random(N)
        true_anomaly = 360.*random.random(N)-180.
        inclination = 180*random.random(N)
        longitude_of_the_ascending_node = 360*random.random(N)-180
        argument_of_periapsis = 360*random.random(N)-180

        comets = datamodel.Particles(N)
        for i, arg in enumerate(zip(mass1, mass2, semi_major_axis, eccentricity, true_anomaly, inclination,
                                   longitude_of_the_ascending_node, argument_of_periapsis)):
            sun_and_comet = new_binary_from_orbital_elements(*arg, G=nbody_system.G)
            comets[i].mass = sun_and_comet[1].mass
            comets[i].position = sun_and_comet[1].position
            comets[i].velocity = sun_and_comet[1].velocity

        semi_major_axis_ext, eccentricity_ext, ta_ext, inclination_ext, \
        longitude_of_the_ascending_node_ext, argument_of_periapsis_ext = \
        orbital_elements_for_rel_posvel_arrays(comets.position,
                                               comets.velocity,
                                               comets.mass + mass_sun,
                                               G=nbody_system.G)
        self.assertAlmostEqual(semi_major_axis, semi_major_axis_ext)
        self.assertAlmostEqual(eccentricity, eccentricity_ext)
        self.assertAlmostEqual(inclination, inclination_ext)
        self.assertAlmostEqual(longitude_of_the_ascending_node, longitude_of_the_ascending_node_ext)
        self.assertAlmostEqual(argument_of_periapsis, argument_of_periapsis_ext)
        self.assertAlmostEqual(true_anomaly, ta_ext)

    def xtest10(self):
        """
        testing orbital_elements_for_rel_posvel_arrays for N particles 
        with random orbital elements, unitless
        """
        numpy.random.seed(666)
        N = 100

        mass_sun = 1.
        mass1 = numpy.ones(N) * mass_sun
        mass2 = numpy.zeros(N)
        semi_major_axis = (-numpy.log(random.random(N)))
        eccentricity = random.random(N)
        true_anomaly = 360.*random.random(N)-180.
        inclination = 180*random.random(N)
        longitude_of_the_ascending_node = 360*random.random(N)-180
        argument_of_periapsis = 360*random.random(N)-180

        comets = datamodel.Particles(N)
        for i, arg in enumerate(zip(mass1, mass2, semi_major_axis, eccentricity, true_anomaly, inclination,
                                   longitude_of_the_ascending_node, argument_of_periapsis)):
            sun_and_comet = new_binary_from_orbital_elements(*arg, G=1)
            comets[i].mass = sun_and_comet[1].mass
            comets[i].position = sun_and_comet[1].position
            comets[i].velocity = sun_and_comet[1].velocity

        semi_major_axis_ext, eccentricity_ext, ta_ext, inclination_ext, \
        longitude_of_the_ascending_node_ext, argument_of_periapsis_ext = \
        orbital_elements_for_rel_posvel_arrays(comets.position,
                                               comets.velocity,
                                               comets.mass + mass_sun,
                                               G=1)

        self.assertAlmostEqual(semi_major_axis, semi_major_axis_ext)
        self.assertAlmostEqual(eccentricity, eccentricity_ext)
        self.assertAlmostEqual(inclination, inclination_ext)
        self.assertAlmostEqual(longitude_of_the_ascending_node, longitude_of_the_ascending_node_ext)
        self.assertAlmostEqual(argument_of_periapsis, argument_of_periapsis_ext)
        self.assertAlmostEqual(true_anomaly, ta_ext)

    def test11(self):
        """
        testing orbital_elements_for_rel_posvel_arrays for unbound orbits
        """

        from amuse.community.kepler.interface import Kepler

        numpy.random.seed(66)
        N = 10

        mass_sun = 1. | units.MSun
        mass1 = numpy.ones(N) * mass_sun
        mass2 = numpy.zeros(N) | units.MSun
        semi_major_axis = -1000.*(random.random(N)) | units.AU
        eccentricity = (1.+random.random(N))*10.-9.
        inclination = numpy.pi*random.random(N)
        longitude_of_the_ascending_node = 2.*numpy.pi*random.random(N)-numpy.pi
        argument_of_periapsis = 2.*numpy.pi*random.random(N)-numpy.pi

        # kepler.initialize_from_elements initializes orbits with mean_anomaly=0 and true_anomaly=0
        true_anomaly = 0.*(360.*random.random(N)-180.)

        comets = datamodel.Particles(N)

        converter = nbody_system.nbody_to_si(1 | units.MSun, 1 | units.AU)
        kepler = Kepler(converter)
        kepler.initialize_code()
        for i, arg in enumerate(zip(mass1, mass2, semi_major_axis, eccentricity, true_anomaly, inclination,
                                   longitude_of_the_ascending_node, argument_of_periapsis)):
            kepler.initialize_from_elements(mass=(mass1[i]+mass2[i]),
                                            semi=semi_major_axis[i],
                                            ecc=eccentricity[i])
            ri = kepler.get_separation_vector()
            vi = kepler.get_velocity_vector()

            om = longitude_of_the_ascending_node[i]
            w = argument_of_periapsis[i]
            incl = inclination[i]
            a1 = ([numpy.cos(om), -numpy.sin(om), 0.0], [numpy.sin(om), numpy.cos(om), 0.0], [0.0, 0.0, 1.0])
            a2 = ([1.0, 0.0, 0.0], [0.0, numpy.cos(incl), -numpy.sin(incl)], [0.0, numpy.sin(incl), numpy.cos(incl)])
            a3 = ([numpy.cos(w), -numpy.sin(w), 0.0], [numpy.sin(w), numpy.cos(w), 0.0], [0.0, 0.0, 1.0])
            A = numpy.dot(numpy.dot(a1, a2), a3)

            r_vec = numpy.dot(A, numpy.reshape(ri, (3, 1)))
            v_vec = numpy.dot(A, numpy.reshape(vi, (3, 1)))

            r = (0.0, 0.0, 0.0) | units.AU
            v = (0.0, 0.0, 0.0) | (units.AU / units.day)
            r[0] = r_vec[0]
            r[1] = r_vec[1]
            r[2] = r_vec[2]
            v[0] = v_vec[0]
            v[1] = v_vec[1]
            v[2] = v_vec[2]

            comets[i].mass = mass2[i]
            comets[i].position = r_vec
            comets[i].velocity = v_vec

        kepler.stop()

        (
            semi_major_axis_ext, eccentricity_ext, ta_ext, inclination_ext,
            longitude_of_the_ascending_node_ext, argument_of_periapsis_ext
        ) = orbital_elements(
            comets.position, comets.velocity, comets.mass + mass_sun,
            G=constants.G,
        )

        self.assertAlmostEqual(semi_major_axis, semi_major_axis_ext.in_(units.au))
        self.assertAlmostEqual(eccentricity, eccentricity_ext)
        self.assertAlmostEqual(inclination, inclination_ext)
        self.assertAlmostEqual(longitude_of_the_ascending_node, longitude_of_the_ascending_node_ext)
        self.assertAlmostEqual(argument_of_periapsis, argument_of_periapsis_ext)
        self.assertAlmostEqual(true_anomaly, ta_ext)

        orbital_parameters = orbital_elements(
            comets.position, comets.velocity, comets.mass + mass_sun,
            G=constants.G, return_dict=True
        )
        self.assertAlmostEqual(
            semi_major_axis,
            orbital_parameters["semi_major_axis"].in_(units.au)
        )
        self.assertAlmostEqual(
            eccentricity,
            orbital_parameters["eccentricity"]
        )
        self.assertAlmostEqual(
            inclination,
            orbital_parameters["inclination"]
        )
        self.assertAlmostEqual(
            longitude_of_the_ascending_node,
            orbital_parameters["longitude_of_the_ascending_node"]
        )
        self.assertAlmostEqual(
            argument_of_periapsis,
            orbital_parameters["argument_of_periapsis"]
        )
        self.assertAlmostEqual(
            true_anomaly,
            orbital_parameters["true_anomaly"]
        )


    def test12(self):
        """
        tests generating cartesian coordinates from orbital elements
        """
        numpy.random.seed(1701)

        mass1 = 1.0 | units.MSun
        mass2 = 0.1 | units.MEarth
        sem = 2. | units.AU
        ecc = 0.15
        inc = 11. | units.deg
        lon = 30. | units.deg
        arg = 0.3 | units.deg
        ta = (360.*random.random()-180.) | units.deg

        rel_pos, rel_vel = rel_posvel_arrays_from_orbital_elements(
                mass1,
                mass2,
                sem,
                ecc,
                ta,
                inc,
                lon,
                arg,
                G=constants.G)

        mass_12 = mass1 + mass2
        sem_ext, ecc_ext, ta_ext, inc_ext, lon_ext, arg_ext = \
            orbital_elements(
                    rel_pos, rel_vel, mass_12, G=constants.G)

        self.assertAlmostEqual(sem, sem_ext)
        self.assertAlmostEqual(ecc, ecc_ext)
        self.assertAlmostEqual(inc, inc_ext)
        self.assertAlmostEqual(lon, lon_ext)
        self.assertAlmostEqual(arg, arg_ext)
        self.assertAlmostEqual(ta, ta_ext)

    def test13(self):
        """
        tests generating cartesian coordinates from orbital elements
        """
        numpy.random.seed(17014)
        N = 5

        mass1 = 1.0 | units.MSun
        mass2 = numpy.ones(N) * 0.01 | units.MEarth
        sem = 2. | units.AU
        ecc = 0.15
        inc = 11. | units.deg
        lon = 30. | units.deg
        arg = 0.3 | units.deg
        ta = (360.*random.random()-180.) | units.deg

        rel_pos, rel_vel = rel_posvel_arrays_from_orbital_elements(
                mass1,
                mass2,
                sem,
                ecc,
                ta,
                inc,
                lon,
                arg,
                G=constants.G)

        mass_12 = mass1 + mass2
        sem_ext, ecc_ext, ta_ext, inc_ext, lon_ext, arg_ext = \
            orbital_elements(
                    rel_pos, rel_vel, mass_12, G=constants.G)

        self.assertAlmostEqual(
                sem, sem_ext)
        self.assertAlmostEqual(ecc, ecc_ext)
        self.assertAlmostEqual(inc, inc_ext)
        self.assertAlmostEqual(lon, lon_ext)
        self.assertAlmostEqual(arg, arg_ext)
        self.assertAlmostEqual(ta, ta_ext)

    def test14(self):
        """
        tests generating cartesian coordinates from orbital elements
        """
        numpy.random.seed(17018)
        N = 5

        mass1 = numpy.ones(N) * 1.0 | units.MSun
        mass2 = random.random(N) | units.MEarth
        sem = numpy.array([2., 1.0, 1.1, 1.2, 4.0]) | units.AU
        ecc = numpy.array([0.15, 0.01, 0.5, 0.9, 0.99])
        inc = numpy.array([11., 0.1, 20, 90, 180.]) | units.deg
        lon = numpy.array([31., 32., 33., 45., 30.]) | units.deg
        arg = numpy.array([0.3, 11., 15., 30., 95.]) | units.deg
        ta = (360.*random.random(N)-180.) | units.deg

        rel_pos, rel_vel = rel_posvel_arrays_from_orbital_elements(
                mass1,
                mass2,
                sem,
                ecc,
                ta,
                inc,
                lon,
                arg,
                G=constants.G)

        mass_12 = mass1 + mass2
        sem_ext, ecc_ext, ta_ext, inc_ext, lon_ext, arg_ext = \
            orbital_elements(
                    rel_pos, rel_vel, mass_12, G=constants.G)
        self.assertAlmostEqual(
                sem.value_in(units.AU), sem_ext.value_in(units.AU))
        self.assertAlmostEqual(ecc, ecc_ext)
        self.assertAlmostEqual(inc, inc_ext)
        self.assertAlmostEqual(lon, lon_ext)
        self.assertAlmostEqual(arg, arg_ext)
        self.assertAlmostEqual(ta, ta_ext)

    def test15(self):
        """
        testing orbital_elements_for_rel_posvel_arrays for N particles
        with random orbital elements
        """
        numpy.random.seed(666)
        N = 100

        mass_sun = 1. | units.MSun
        mass1 = numpy.ones(N) * mass_sun
        mass2 = numpy.zeros(N) | units.MSun
        semi_major_axis = (-numpy.log(random.random(N))) | units.AU
        eccentricity = random.random(N)
        true_anomaly = 360.*random.random(N)-180.
        inclination = 180*random.random(N)
        longitude_of_the_ascending_node = 360*random.random(N)-180
        argument_of_periapsis = 360*random.random(N)-180

        comets = datamodel.Particles(N)
        suns = datamodel.Particles(N)
        for i, arg in enumerate(
                zip(mass1, mass2, semi_major_axis, eccentricity, true_anomaly,
                    inclination, longitude_of_the_ascending_node,
                    argument_of_periapsis)):
            sun_and_comet = new_binary_from_orbital_elements(
                    *arg, G=constants.G)
            comets[i].mass = sun_and_comet[1].mass
            comets[i].position = sun_and_comet[1].position
            comets[i].velocity = sun_and_comet[1].velocity

        suns.mass = mass1
        suns.position = 0*comets.position
        suns.velocity = 0*comets.velocity

        mass1_ext, mass2_ext, semi_major_axis_ext, eccentricity_ext, ta_ext,\
            inclination_ext, longitude_of_the_ascending_node_ext,\
            argument_of_periapsis_ext = orbital_elements(
                    suns, comets, G=constants.G)
        rad_to_deg = 180./numpy.pi
        for i in range(N):
            self.assertAlmostEqual(
                    semi_major_axis[i].value_in(units.AU),
                    semi_major_axis_ext[i].value_in(units.AU))
            self.assertAlmostEqual(eccentricity[i], eccentricity_ext[i])
            self.assertAlmostEqual(
                    inclination[i], rad_to_deg*inclination_ext[i])
            self.assertAlmostEqual(
                    longitude_of_the_ascending_node[i],
                    rad_to_deg*longitude_of_the_ascending_node_ext[i])
            self.assertAlmostEqual(
                    argument_of_periapsis[i],
                    rad_to_deg*argument_of_periapsis_ext[i])
            self.assertAlmostEqual(true_anomaly[i], rad_to_deg*ta_ext[i])

    def test16(self):
        """ tests a mismatch in shape in generate_binaries """
        m1 = [1]*5 | nbody_system.mass
        m2 = [0]*5 | nbody_system.mass
        a = [1.]*5 | nbody_system.length
        ecc = numpy.array([0, 0, .99999, 0.1, 0.5])
        ta = [180, 180, 20, 30, 0] | units.deg
        primaries, secondaries = generate_binaries(m1, m2, a, eccentricity=ecc, true_anomaly=ta)
        m1_, m2_, a_, ecc_, ta_, i_, lasc_, ap_ = get_orbital_elements_from_binaries(primaries, secondaries)
        self.assertAlmostEqual(ecc, ecc_)



from amuse.test.amusetest import TestWithMPI, get_path_to_results
import sys
import os.path
from subprocess import PIPE, Popen
import numpy
from math import ceil

from amuse.community.mesa_r2208.interface import MESA, MESAInterface

from amuse.support.exceptions import AmuseException
from amuse.units import units
from amuse.datamodel import Particles
from amuse.datamodel import Particle
from amuse.rfi import channel
from amuse.ext.spherical_model import EnclosedMassInterpolator


class TestMESAInterface(TestWithMPI):

    def test1(self):
        print("Testing initialization of the interface...")
        instance = self.new_instance_of_an_optional_code(MESAInterface)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        inlist_path = instance.default_path_to_inlist
        # print "Path to inlist: ", inlist_path
        MESA_data_path = instance.default_path_to_MESA_data
        # print "Path to MESA data directory: ", MESA_data_path
        instance.set_MESA_paths(instance.default_path_to_inlist,
            instance.default_path_to_MESA_data, instance.get_output_directory())
        status = instance.initialize_code()
        self.assertEqual(status, 0)
        instance.stop()

    def slowtest2(self):
        print("Testing get/set of metallicity (tests new ZAMS model implicitly)...")
        print("The first time this test will take quite some time"
            " to generate new starting models.")
        instance = self.new_instance_of_an_optional_code(MESAInterface)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        instance.set_MESA_paths(instance.default_path_to_inlist,
            instance.default_path_to_MESA_data, instance.get_output_directory())
        status = instance.initialize_code()
        self.assertEqual(status, 0)
        (metallicity, error) = instance.get_metallicity()
        self.assertEqual(0, error)
        self.assertEqual(0.02, metallicity)
        for x in [0.04, 0.02, 0.01, 0.00]:
            error = instance.set_metallicity(x)
            self.assertEqual(0, error)
            (metallicity, error) = instance.get_metallicity()
            self.assertEqual(0, error)
            self.assertEqual(x, metallicity)
        instance.stop()

    def test3(self):
        print("Testing basic operations: new_particle...")
        instance = self.new_instance_of_an_optional_code(MESAInterface)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        (maximum_number_of_stars, error) = instance.get_maximum_number_of_stars()
        self.assertEqual(0, error)
        self.assertEqual(maximum_number_of_stars, 1000)
        instance.set_MESA_paths(instance.default_path_to_inlist,
            instance.default_path_to_MESA_data, instance.get_output_directory())
        status = instance.initialize_code()
        self.assertEqual(status, 0)
        number_of_stars = 1
        for i in range(number_of_stars):
            # print i
            (index_of_the_star, error) = instance.new_particle(0.5+i*1.0/number_of_stars)
            self.assertEqual(0, error)
            self.assertEqual(index_of_the_star, i+1)
        # import time    # To check the amount of memory is used ...
        # time.sleep(10) # when a large number of stars is created.
        instance.stop()

    def test4(self):
        print("Testing basic operations: evolve...")
        instance = self.new_instance_of_an_optional_code(MESAInterface)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        instance.set_MESA_paths(instance.default_path_to_inlist,
            instance.default_path_to_MESA_data, instance.get_output_directory())
        status = instance.initialize_code()
        (index_of_the_star, error) = instance.new_particle(1.0)
        self.assertEqual(0, error)
        self.assertEqual(index_of_the_star, 1)
        self.assertEqual(0, instance.commit_particles())

        initial_dt = 1.0e5
        dt_factor = 1.2
        self.assertEqual([initial_dt, 0], list(instance.get_time_step(index_of_the_star).values()))
        self.assertEqual(0, instance.evolve_one_step(index_of_the_star))
        self.assertEqual([initial_dt, 0], list(instance.get_age(index_of_the_star).values()))

        target_end_time = 3.0e5  # (years)
        self.assertEqual(0, instance.evolve_for(index_of_the_star, target_end_time-initial_dt))
        self.assertEqual([initial_dt*(1 + dt_factor + dt_factor**2), 0], list(instance.get_age(index_of_the_star).values()))
        self.assertEqual([round(initial_dt*dt_factor**3), 0], list(instance.get_time_step(index_of_the_star).values()))
        self.assertTrue(instance.get_age(index_of_the_star)['age'] >= target_end_time)

        (L_of_the_star, error) = instance.get_luminosity(index_of_the_star)
        self.assertEqual(0, error)
        self.assertAlmostEqual(L_of_the_star, 0.725, 1)
        (M_of_the_star, error) = instance.get_mass(index_of_the_star)
        self.assertEqual(0, error)
        self.assertAlmostEqual(M_of_the_star, 1.000, 3)
        (T_of_the_star, error) = instance.get_temperature(index_of_the_star)
        self.assertEqual(0, error)
        self.assertAlmostEqual(T_of_the_star, 5650.998, -2)
        instance.stop()

    def slowtest5(self):
        print("Testing evolve with varying Z (tests new ZAMS model implicitly)...")
        print("If the required starting models do not exist, this test will "
            "take quite some time to generate them.")
        instance = self.new_instance_of_an_optional_code(MESAInterface)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        instance.set_MESA_paths(instance.default_path_to_inlist,
            instance.default_path_to_MESA_data, instance.get_output_directory())
        status = instance.initialize_code()
        self.assertEqual(status, 0)
        metallicities = [0.00, 0.01, 0.02, 0.04]
        luminosities = [1.717, 0.938, 0.725, 0.592]
        for (i, (Z, L)) in enumerate(zip(metallicities, luminosities)):
            status = instance.set_metallicity(Z)
            self.assertEqual(0, status)
            (index_of_the_star, status) = instance.new_particle(1.0)
            self.assertEqual(0, status)
            self.assertEqual(index_of_the_star, i+1)
            instance.evolve_for(index_of_the_star, 5.0e5)
            (L_of_the_star, status) = instance.get_luminosity(index_of_the_star)
            self.assertEqual(0, status)
            self.assertAlmostEqual(L_of_the_star, L, 1)
        instance.stop()

    def test6(self):
        print("Testing MESA stop conditions...")
        instance = self.new_instance_of_an_optional_code(MESAInterface)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        (value, error) = instance.get_max_age_stop_condition()
        self.assertEqual(0, error)
        self.assertEqual(1.0e36, value)
        for x in range(10, 14):
            error = instance.set_max_age_stop_condition(10 ** x)
            self.assertEqual(0, error)
            (value, error) = instance.get_max_age_stop_condition()
            self.assertEqual(0, error)
            self.assertEqual(10 ** x, value)
        (value, error) = instance.get_min_timestep_stop_condition()
        self.assertEqual(0, error)
        self.assertEqual(1.0e-6, value)
        for x in range(-9, -2):
            error = instance.set_min_timestep_stop_condition(10.0 ** x)
            self.assertEqual(0, error)
            (value, error) = instance.get_min_timestep_stop_condition()
            self.assertEqual(0, error)
            self.assertEqual(10.0 ** x, value)
        instance.stop()

    def test7(self):
        print("Testing MESA parameters...")
        instance = self.new_instance_of_an_optional_code(MESAInterface)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        (value, error) = instance.get_mixing_length_ratio()
        self.assertEqual(0, error)
        self.assertEqual(2.0, value)
        for x in [1.0, 1.5, 3.0]:
            error = instance.set_mixing_length_ratio(x)
            self.assertEqual(0, error)
            (value, error) = instance.get_mixing_length_ratio()
            self.assertEqual(0, error)
            self.assertEqual(x, value)
        (value, error) = instance.get_semi_convection_efficiency()
        self.assertEqual(0, error)
        self.assertEqual(0.0, value)
        for x in [0.1, 0.04, 0.001]:
            error = instance.set_semi_convection_efficiency(x)
            self.assertEqual(0, error)
            (value, error) = instance.get_semi_convection_efficiency()
            self.assertEqual(0, error)
            self.assertEqual(x, value)
        instance.stop()

    def test8(self):
        print("Testing MESA wind parameters...")
        instance = self.new_instance_of_an_optional_code(MESAInterface)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        (value, error) = instance.get_RGB_wind_scheme()
        self.assertEqual(0, error)
        self.assertEqual(1, value)
        for x in range(6):
            error = instance.set_RGB_wind_scheme(x)
            self.assertEqual(0, error)
            (value, error) = instance.get_RGB_wind_scheme()
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        (value, error) = instance.get_AGB_wind_scheme()
        self.assertEqual(0, error)
        self.assertEqual(1, value)
        for x in range(6):
            error = instance.set_AGB_wind_scheme(x)
            self.assertEqual(0, error)
            (value, error) = instance.get_AGB_wind_scheme()
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        (value, error) = instance.get_reimers_wind_efficiency()
        self.assertEqual(0, error)
        self.assertEqual(0.5, value)
        for x in [0.0, 0.1, 0.5, 1.0]:
            error = instance.set_reimers_wind_efficiency(x)
            self.assertEqual(0, error)
            (value, error) = instance.get_reimers_wind_efficiency()
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        (value, error) = instance.get_blocker_wind_efficiency()
        self.assertEqual(0, error)
        self.assertEqual(0.1, value)
        for x in [0.0, 0.1, 0.5, 1.0]:
            error = instance.set_blocker_wind_efficiency(x)
            self.assertEqual(0, error)
            (value, error) = instance.get_blocker_wind_efficiency()
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        (value, error) = instance.get_de_jager_wind_efficiency()
        self.assertEqual(0, error)
        self.assertEqual(0.8, value)
        for x in [0.0, 0.1, 0.5, 1.0]:
            error = instance.set_de_jager_wind_efficiency(x)
            self.assertEqual(0, error)
            (value, error) = instance.get_de_jager_wind_efficiency()
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        (value, error) = instance.get_dutch_wind_efficiency()
        self.assertEqual(0, error)
        self.assertEqual(0.8, value)
        for x in [0.0, 0.1, 0.5, 1.0]:
            error = instance.set_dutch_wind_efficiency(x)
            self.assertEqual(0, error)
            (value, error) = instance.get_dutch_wind_efficiency()
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        instance.stop()


class TestMESA(TestWithMPI):

    def test1(self):
        print("Testing initialization and default MESA parameters...")
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        instance.set_MESA_paths(instance.default_path_to_inlist,
            instance.default_path_to_MESA_data, instance.get_output_directory())
        instance.initialize_code()
        self.assertEqual(0.02 | units.no_unit, instance.parameters.metallicity)
        self.assertEqual(1.0e36 | units.yr, instance.parameters.max_age_stop_condition)
        instance.parameters.max_age_stop_condition = 1.0e2 | units.Myr
        self.assertEqual(1.0e2 | units.Myr, instance.parameters.max_age_stop_condition)
        instance.stop()

    def test2(self):
        print("Testing basic operations: evolve and get_...")
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        instance.initialize_code()
        instance.commit_parameters()
        index_of_the_star = instance.new_particle(1.0 | units.MSun)
        self.assertEqual(index_of_the_star, 1)

        initial_dt = 1.0e5 | units.yr
        dt_factor = 1.2
        time_step = instance.get_time_step(index_of_the_star)
        self.assertAlmostEqual(time_step, initial_dt)
        instance.evolve_one_step(index_of_the_star)
        age_of_the_star = instance.get_age(index_of_the_star)
        self.assertAlmostEqual(age_of_the_star, initial_dt)

        target_end_time = 3.0e5 | units.yr
        instance.evolve_for(index_of_the_star, target_end_time - age_of_the_star)
        self.assertAlmostEqual(initial_dt*(1 + dt_factor + dt_factor**2), instance.get_age(index_of_the_star))
        self.assertAlmostEqual(initial_dt*dt_factor**3, instance.get_time_step(index_of_the_star))
        self.assertTrue(instance.get_age(index_of_the_star) >= target_end_time)

        L_of_the_star = instance.get_luminosity(index_of_the_star)
        self.assertAlmostEqual(L_of_the_star, 0.725 | units.LSun, 1)
        M_of_the_star = instance.get_mass(index_of_the_star)
        self.assertAlmostEqual(M_of_the_star, 1.000 | units.MSun, 3)
        T_of_the_star = instance.get_temperature(index_of_the_star)
        self.assertAlmostEqual(T_of_the_star, 5650.998 | units.K, -2)
        instance.stop()

    def test3(self):
        print("Testing basic operations: evolve_model and channels...")
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        instance.initialize_code()
        instance.commit_parameters()
        stars = Particles(1)
        mass = 10. | units.MSun
        stars.mass = mass
        instance.particles.add_particles(stars)
        instance.commit_particles()
        from_code_to_model = instance.particles.new_channel_to(stars)
        from_code_to_model.copy()
        # print stars
        # instance.evolve_model(end_time = 0.03 | units.Myr) # speeding test up:
        self.assertEqual(stars[0].mass, mass)
        self.assertAlmostRelativeEqual(stars[0].luminosity, 5841. | units.LSun, 1)
        instance.evolve_model()
        from_code_to_model.copy()
        self.assertAlmostEqual(stars[0].mass, mass, 5)
        self.assertAlmostRelativeEqual(stars[0].luminosity, 5820.85 | units.LSun, 1)
        instance.stop()

    def slowtest4(self):
        print("Testing stellar type...")
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        instance.initialize_code()
        instance.commit_parameters()
        stars = Particles(1)

        star = stars[0]
        star.mass = 5.0 | units.MSun

        instance.particles.add_particles(stars)
        instance.commit_particles()

        from_code_to_model = instance.particles.new_channel_to(stars)
        from_code_to_model.copy()

        previous_type = 15 | units.stellar_type
        results = []
        current_time = 0 | units.Myr

        while current_time < (101 | units.Myr):
            if not star.stellar_type == previous_type:
                print((star.age, star.mass, star.stellar_type))
                results.append((star.age, star.mass, star.stellar_type))
                previous_type = star.stellar_type
            instance.evolve_model()
            from_code_to_model.copy()
            current_time = star.age

        self.assertEqual(len(results), 4)

        times = (
            0.0 | units.Myr,
            81.6 | units.Myr,
            99.9 | units.Myr,
            100.3 | units.Myr
        )
        for result, expected in zip(results, times):
            self.assertAlmostEqual(result[0].value_in(units.Myr), expected.value_in(units.Myr), 1)

        masses = (
            5.000 | units.MSun,
            5.000 | units.MSun,
            5.000 | units.MSun,
            5.000 | units.MSun
        )
        for result, expected in zip(results, masses):
            self.assertAlmostEqual(result[1].value_in(units.MSun), expected.value_in(units.MSun), 3)

        types = (
            "Main Sequence star",
            "First Giant Branch",
            "Core Helium Burning",
            "First Asymptotic Giant Branch"
        )

        for result, expected in zip(results, types):
            self.assertEqual(str(result[2]), expected)

        instance.stop()

    def test5(self):
        print("Testing evolve_model for particle set...")
        instance = self.new_instance_of_an_optional_code(MESA)
        masses = [0.5, 1.0] | units.MSun
        max_age = 0.6 | units.Myr
        number_of_stars = len(masses)
        stars = Particles(number_of_stars)
        stars.mass = masses
        instance.initialize_code()
        self.assertEqual(instance.parameters.max_age_stop_condition, 1e30 | units.Myr)
        instance.parameters.max_age_stop_condition = max_age
        self.assertEqual(instance.parameters.max_age_stop_condition, max_age)
        instance.particles.add_particles(stars)
        instance.commit_particles()
        from_code_to_model = instance.particles.new_channel_to(stars)
        instance.evolve_model(end_time=0.5 | units.Myr)
        from_code_to_model.copy()
        # print stars
        for i in range(number_of_stars):
            self.assertTrue(stars[i].age.value_in(units.Myr) >= 0.5)
            self.assertTrue(stars[i].age <= max_age)
            self.assertTrue(stars[i].mass <= masses[i])
            self.assertTrue(stars[i].age+stars[i].time_step <= max_age)

        self.assertRaises(AmuseException, instance.evolve_model, end_time=2*max_age,
            expected_message="Error when calling 'evolve_for' of a 'MESA', "
                "errorcode is -12, error is 'Evolve terminated: Maximum age reached.'")
        instance.stop()

    def test6(self):
        print("Test for obtaining the stellar structure model")
        stars = Particles(2)
        stars.mass = [1.0, 10.0] | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(stars)
        instance.commit_particles()
        instance.evolve_model()
        self.assertEqual(instance.particles.get_number_of_zones(), [575, 2262])
        self.assertEqual(len(instance.particles[0].get_mass_profile()), 575)
        self.assertAlmostEqual(instance.particles[0].get_mass_profile().sum(), 1.0)
        self.assertRaises(AmuseException, instance.particles.get_mass_profile,
            expected_message="Querying mass profiles of more than one particle at a time is not supported.")
        print(instance.particles)
        self.assertEqual(len(instance.particles[1].get_density_profile()), 2262)
        self.assertIsOfOrder(instance.particles[0].get_radius_profile()[-1],          1.0 | units.RSun)
        self.assertIsOfOrder(instance.particles[0].get_temperature_profile()[0],  1.0e7 | units.K)
        self.assertIsOfOrder(instance.particles[0].get_luminosity_profile()[-1],      1.0 | units.LSun)
        self.assertIsOfOrder(instance.particles[0].get_pressure_profile()[0],      1.0e17 | units.barye)
        delta_mass = instance.particles[0].get_mass_profile() * instance.particles[0].mass
        radius1 = instance.particles[0].get_radius_profile()
        radius2 = radius1[:-1]
        radius2.prepend(0 | units.m)
        delta_radius_cubed = (radius1**3 - radius2**3)
        self.assertAlmostEqual(instance.particles[0].get_density_profile() /
            (delta_mass/(4./3.*numpy.pi*delta_radius_cubed)), [1]*575, places=3)
        self.assertAlmostEqual(instance.particles[1].get_mu_profile(), [0.62]*2262 | units.amu, places=1)
        instance.stop()

    def test7(self):
        print("Test for obtaining the stellar composition structure")
        stars = Particles(1)
        stars.mass = 1.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(stars)
        instance.commit_particles()
        instance.evolve_model()
        number_of_zones = instance.particles.get_number_of_zones()[0]
        number_of_species = instance.particles.get_number_of_species()[0]
        composition = instance.particles[0].get_chemical_abundance_profiles()
        species_names = instance.particles[0].get_names_of_species()
        species_IDs = instance.particles[0].get_IDs_of_species()
        species_masses = instance.particles[0].get_masses_of_species()
        self.assertEqual(number_of_zones,    575)
        self.assertEqual(number_of_species,    8)
        self.assertEqual(len(species_names),  number_of_species)
        self.assertEqual(len(composition),    number_of_species)
        self.assertEqual(len(composition[0]), number_of_zones)
        self.assertEqual(species_names, ['h1', 'he3', 'he4', 'c12', 'n14', 'o16', 'ne20', 'mg24'])
        self.assertEqual(species_IDs,   [2,    5,     6,     38,    51,    69,    114,    168])
        self.assertAlmostEqual(species_masses, [1.0078250, 3.0160293, 4.0026032, 12.0,
                                14.0030740, 15.9949146, 19.9924401, 23.9850417] | units.amu, places=5)
        self.assertAlmostEqual(composition[:1, -1].sum(),  0.7)
        self.assertAlmostEqual(composition[1:3, -1].sum(),  (0.3) - instance.parameters.metallicity)
        self.assertAlmostEqual(composition[3:, -1].sum(),  instance.parameters.metallicity)
        self.assertAlmostEqual(composition.sum(axis=0), [1.0]*number_of_zones)
        instance.stop()

    def slowtest8(self):
        print("Test for obtaining the stellar composition structure - evolved star with zero metalicity")
        stars = Particles(1)
        stars.mass = 1.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        instance.initialize_code()
        instance.parameters.metallicity = 0.0
        instance.commit_parameters()
        instance.particles.add_particles(stars)
        instance.commit_particles()
        instance.evolve_model(5.85 | units.Gyr)
        self.assertTrue(instance.particles[0].age >= 5.85 | units.Gyr)
        self.assertTrue(str(instance.particles[0].stellar_type) == "First Giant Branch")
        number_of_zones = instance.particles.get_number_of_zones()[0]
        number_of_species = instance.particles.get_number_of_species()[0]
        composition = instance.particles[0].get_chemical_abundance_profiles()
        species_names = instance.particles[0].get_names_of_species()
        species_IDs = instance.particles[0].get_IDs_of_species()
        self.assertEqual(number_of_zones,    578)
        self.assertEqual(number_of_species,    8)
        self.assertEqual(len(species_names),  number_of_species)
        self.assertEqual(len(composition),    number_of_species)
        self.assertEqual(len(composition[0]), number_of_zones)
        self.assertEqual(species_names, ['h1', 'he3', 'he4', 'c12', 'n14', 'o16', 'ne20', 'mg24'])
        self.assertEqual(species_IDs,   [2,    5,     6,     38,    51,    69,    114,    168])
        self.assertAlmostEqual(composition[:1, number_of_zones-1].sum(),  0.76 | units.none)
        self.assertAlmostEqual(composition[1:3, number_of_zones-1].sum(),  0.24 | units.none)
        self.assertAlmostEqual(composition[3:, number_of_zones-1].sum(),  0.00 | units.none)
        self.assertAlmostEqual(composition.sum(axis=0), [1.0]*number_of_zones | units.none)
        self.assertAlmostEqual(composition[:1, 0].sum(),  0.00 | units.none)
        self.assertAlmostEqual(composition[1:3, 0].sum(),  1.00 | units.none)
        self.assertAlmostEqual(composition[3:, 0].sum(),  0.00 | units.none)
        instance.stop()

    def test9(self):
        print("Test for changing the stellar structure model")
        star = Particles(1)
        star.mass = 1.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(star)
        instance.commit_particles()
        instance.evolve_model()

        density_profile = instance.particles[0].get_density_profile()

        self.assertRaises(AmuseException, instance.particles[0].set_density_profile, density_profile[2:],
            expected_message="The length of the supplied vector (573) does not match the number of "
            "mesh zones of the star (575).")

        mass_factor = 1.1
        instance.particles[0].set_density_profile(mass_factor*density_profile)
        self.assertAlmostRelativeEqual(instance.particles[0].get_density_profile(), density_profile*mass_factor, places=10)
        instance.particles.mass *= mass_factor
        instance.evolve_model()

        outer_radius = instance.particles[0].get_radius_profile()
        inner_radius = outer_radius[:-1]
        inner_radius.prepend(0 | units.m)
        delta_radius_cubed = (outer_radius**3 - inner_radius**3)
        integrated_mass = (4./3.*numpy.pi*delta_radius_cubed*instance.particles[0].get_density_profile()).sum()
        self.assertAlmostRelativeEqual(integrated_mass, star.mass*mass_factor, places=3)
        instance.stop()

    def test10(self):
        print("Test for changing the stellar composition")
        star = Particles(1)
        star.mass = 1.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)  # , redirection = 'none')
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(star)
        instance.commit_particles()
        instance.evolve_model(0.3 | units.Myr)

        composition = instance.particles[0].get_chemical_abundance_profiles()
        k_surface = -1  # index to the outer mesh cell (surface)

        self.assertAlmostEqual(composition[:1, k_surface].sum(),  0.7)
        self.assertAlmostEqual(composition[1:3, k_surface].sum(),  (0.3) - instance.parameters.metallicity)
        self.assertAlmostEqual(composition[3:, k_surface].sum(),  instance.parameters.metallicity)

        # Gradually and consistently increase helium and decrease hydrogen abundances until reversed
        for alpha in [0.3, 1.0, -0.5, -0.125]:
            h1_profile = composition[0] * 1
            he4_profile = composition[2] * 1
            composition[0] = alpha * he4_profile + (1-alpha) * h1_profile
            composition[2] = (1-alpha) * he4_profile + alpha * h1_profile
            instance.particles[0].set_chemical_abundance_profiles(composition)
            instance.evolve_model()
            instance.evolve_model()
            composition = instance.particles[0].get_chemical_abundance_profiles()

        self.assertAlmostEqual(composition[:2, k_surface].sum(),  (0.3) - instance.parameters.metallicity)
        self.assertAlmostEqual(composition[2:3, k_surface].sum(),  0.7)
        self.assertAlmostEqual(composition[3:, k_surface].sum(),  instance.parameters.metallicity)
        self.assertAlmostEqual(composition.sum(axis=0), 1.0)

        self.assertRaises(AmuseException, instance.particles[0].set_chemical_abundance_profiles, composition[:7],
            expected_message="The length of the supplied vector (7) does not match the number of "
            "chemical species of the star (8).")
        instance.stop()

    def test11(self):
        print("Test evolve_model optional arguments: end_time and keep_synchronous")
        stars = Particles(3)
        stars.mass = [1.0, 2.0, 3.0] | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(stars)
        instance.commit_particles()

        self.assertAlmostEqual(instance.particles.age, [0.0, 0.0, 0.0] | units.yr)
        self.assertAlmostEqual(instance.particles.time_step, [100000.0, 17677.6695, 6415.0029] | units.yr, 3)

        print("evolve_model without arguments: use shared timestep = 0.99*min(particles.time_step)")
        instance.evolve_model()
        self.assertAlmostEqual(instance.particles.age, [100000.0, 17677.6695, 6415.0029] | units.yr, 3)
        self.assertAlmostRelativeEquals(instance.particles.time_step, 1.2*([100000.0, 17677.6695, 6415.0029] | units.yr), 6)
        self.assertAlmostEqual(instance.model_time, 0.99 * 6415.0029 | units.yr, 3)

        print("evolve_model with end_time: take timesteps, until end_time is reached exactly")
        instance.evolve_model(15000 | units.yr)
        self.assertAlmostEqual(instance.particles.age, [100000.0, 17677.6695, 6415.0029*(1+1.2+1.44)] | units.yr, 3)
        self.assertAlmostRelativeEquals(instance.particles.time_step, 1.2*([100000.0, 17677.6695, 1.44*6415.0029] | units.yr), 4)
        self.assertAlmostEqual(instance.model_time, 15000.0 | units.yr, 3)

        print("evolve_model with keep_synchronous: use non-shared timestep, particle ages will typically diverge")
        instance.evolve_model(keep_synchronous=False)
        self.assertAlmostRelativeEquals(instance.particles.age, ([100000.0, 17677.6695, 6415.0029*(1+1.2+1.44)] | units.yr)
            + 1.2*([100000.0, 17677.6695, 1.44*6415.0029] | units.yr), 5)
        self.assertAlmostRelativeEquals(instance.particles.time_step, 1.44*([100000.0, 17677.6695, 1.44*6415.0029] | units.yr), 4)
        self.assertAlmostEqual(instance.model_time, 15000.0 | units.yr, 3)  # Unchanged!
        instance.stop()

    def test12(self):
        print("Test for importing new stellar models")
        star = Particles(1)
        star.mass = 1.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        instance.initialize_code()
        instance.parameters.stabilize_new_stellar_model_flag = False
        instance.commit_parameters()
        instance.particles.add_particles(star)
        instance.commit_particles()
        instance.evolve_model()

        number_of_zones = instance.particles[0].get_number_of_zones()
        composition = instance.particles[0].get_chemical_abundance_profiles(number_of_zones=number_of_zones)
        instance.new_particle_from_model(dict(
            mass=instance.particles[0].get_cumulative_mass_profile(number_of_zones=number_of_zones) * instance.particles[0].mass,
            radius=instance.particles[0].get_radius_profile(number_of_zones=number_of_zones),
            rho=instance.particles[0].get_density_profile(number_of_zones=number_of_zones),
            temperature=instance.particles[0].get_temperature_profile(number_of_zones=number_of_zones),
            luminosity=instance.particles[0].get_luminosity_profile(number_of_zones=number_of_zones),
            X_H=composition[0],
            X_He=composition[1] + composition[2],
            X_C=composition[3],
            X_N=composition[4],
            X_O=composition[5],
            X_Ne=composition[6],
            X_Mg=composition[7],
            X_Si=composition[7]*0.0,
            X_Fe=composition[7]*0.0), 10.0 | units.Myr)
        self.assertEqual(len(instance.particles), 2)
        self.assertEqual(len(instance.imported_stars), 1)
        self.assertEqual(instance.imported_stars[0].get_number_of_zones(), number_of_zones)
        self.assertIsOfOrder(instance.imported_stars[0].get_radius_profile()[-1],          1.0 | units.RSun)
        self.assertIsOfOrder(instance.imported_stars[0].get_temperature_profile()[0],  1.0e7 | units.K)
        self.assertIsOfOrder(instance.imported_stars[0].get_pressure_profile()[0],      1.0e17 | units.barye)
        self.assertAlmostEqual(instance.imported_stars[0].get_mass_profile(),
                               instance.native_stars[0].get_mass_profile())
        self.assertAlmostRelativeEqual(instance.imported_stars[0].get_pressure_profile(),
                               instance.native_stars[0].get_pressure_profile(), 7)
        self.assertAlmostEqual(instance.imported_stars[0].get_radius_profile(),
                               instance.native_stars[0].get_radius_profile())
        self.assertAlmostEqual(instance.imported_stars[0].get_temperature_profile(),
                               instance.native_stars[0].get_temperature_profile())

        print(instance.particles)
        instance.evolve_model(keep_synchronous=False)
        print(instance.particles)
        instance.stop()

    def slowtest13(self):
        print("Testing MESA wind parameters...")
        stars = Particles(9)
        stars.mass = 10.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        instance.initialize_code()
        instance.parameters.reimers_wind_efficiency = 0.5
        instance.parameters.blocker_wind_efficiency = 0.1
        instance.parameters.de_jager_wind_efficiency = 0.8
        instance.parameters.dutch_wind_efficiency = 0.8
        instance.commit_parameters()
        for wind_scheme in [0, 1, 2, 3, 4]:
            instance.parameters.RGB_wind_scheme = wind_scheme
            instance.recommit_parameters()
            instance.particles.add_particle(stars[wind_scheme])
        instance.parameters.reimers_wind_efficiency *= 2.0
        instance.parameters.blocker_wind_efficiency *= 2.0
        instance.parameters.de_jager_wind_efficiency *= 2.0
        instance.parameters.dutch_wind_efficiency *= 2.0
        for wind_scheme in [1, 2, 3, 4]:
            instance.parameters.RGB_wind_scheme = wind_scheme
            instance.recommit_parameters()
            instance.particles.add_particle(stars[wind_scheme+4])
        instance.commit_particles()
        instance.evolve_model(keep_synchronous=False)
        from_code_to_model = instance.particles.new_channel_to(stars)
        from_code_to_model.copy()
        print(stars)
        self.assertAlmostEqual(stars[0].wind, 0.0 | units.MSun / units.yr)
        self.assertAlmostRelativeEqual(stars[1:5].wind,
            [4.59318475897e-10, 5.20742729636e-11, 1.05565558121e-09, 3.62519254311e-09] | units.MSun / units.yr, places=7)
        self.assertAlmostRelativeEqual(stars[5:].wind, 2.0 * stars[1:5].wind, places=7)
        instance.stop()

    def test14(self):
        print("Testing MESA wind parameters... (short version of slowtest13)")
        stars = Particles(3)
        stars.mass = 10.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        instance.initialize_code()
        instance.parameters.RGB_wind_scheme = 0
        instance.commit_parameters()
        instance.particles.add_particle(stars[0])
        instance.parameters.RGB_wind_scheme = 1
        for i, wind_efficiency in enumerate([0.5, 1.0]):
            instance.parameters.reimers_wind_efficiency = wind_efficiency
            instance.recommit_parameters()
            instance.particles.add_particle(stars[i+1])
        instance.commit_particles()
        instance.evolve_model(keep_synchronous=False)
        from_code_to_model = instance.particles.new_channel_to(stars)
        from_code_to_model.copy()
        self.assertAlmostEqual(stars[0].wind, 0.0 | units.MSun / units.yr)
        self.assertAlmostRelativeEqual(stars[1].wind, 4.59318475897e-10 | units.MSun / units.yr, places=1)
        self.assertAlmostRelativeEqual(stars[2].wind, 2.0 * stars[1].wind, places=7)
        instance.stop()

    def test15(self):
        print("Testing MESA states")
        stars = Particles(2)
        stars.mass = 1.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return

        print("First do everything manually:", end=' ')
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.initialize_code()
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.commit_parameters()
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        instance.particles.add_particle(stars[0])
        instance.commit_particles()
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.cleanup_code()
        self.assertEqual(instance.get_name_of_current_state(), 'END')
        instance.stop()
        print("ok")

        print("initialize_code(), commit_parameters(), (re)commit_particles(), "
            "and cleanup_code() should be called automatically:", end=' ')
        instance = MESA()
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.parameters.RGB_wind_scheme = 1
        instance.parameters.reimers_wind_efficiency = 0.5
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.particles.add_particle(stars[0])
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        mass = instance.particles[0].mass
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.particles.add_particle(stars[1])
        self.assertEqual(instance.get_name_of_current_state(), 'UPDATE')
        mass = instance.particles[0].mass
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.stop()
        self.assertEqual(instance.get_name_of_current_state(), 'STOPPED')
        print("ok")

    def test16(self):
        print("Testing basic operations: evolve_one_step and evolve_for")
        stars = Particles(2)
        stars.mass = 1.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)  # , redirection = 'none')
        if instance is None:
            print("MESA was not built. Skipping test.")
            return

        se_stars = instance.particles.add_particles(stars)

        for i in range(3):
            se_stars[0].evolve_one_step()
        self.assertAlmostEqual(se_stars.age, [364000.0, 0] | units.yr)

        se_stars[1].evolve_for(se_stars[0].age)
        self.assertAlmostRelativeEqual(se_stars[0].age,         se_stars[1].age)
        self.assertAlmostRelativeEqual(se_stars[0].luminosity,  se_stars[1].luminosity, 2)
        self.assertAlmostRelativeEqual(se_stars[0].radius,      se_stars[1].radius, 2)
        self.assertAlmostRelativeEqual(se_stars[0].temperature, se_stars[1].temperature, 2)
        instance.stop()

    def test17(self):
        print("MESA validation")

        mesa_src_path = os.path.join(os.path.dirname(sys.modules[MESA.__module__].__file__), 'src', 'mesa')
        mesa_star_path = os.path.join(mesa_src_path, 'star', 'test', 'star')

        if not os.path.exists(mesa_star_path) or not os.access(mesa_star_path, os.X_OK):
            self.skip("no mesa executable in binary distribution, test cannot run")

        number_of_steps = 11  # 4
        star = Particle()
        star.mass = 1.0 | units.MSun

        testpath = get_path_to_results()
        outputfile_name = os.path.join(testpath, "mesa_output")

        instance = self.new_instance_of_an_optional_code(MESA, redirection='file', redirect_file=outputfile_name)
        se_star = instance.particles.add_particle(star)
        for i in range(number_of_steps):
            se_star.evolve_one_step()
        instance.stop()

        rfile = open(outputfile_name, 'r')
        amuse_output = rfile.read()
        rfile.close()

        # generate the inlist for the (stand-alone) MESA star run:
        instance = self.new_instance_of_an_optional_code(MESA, redirection='null')
        with open(instance.default_path_to_inlist, 'r') as default_inlist:
            with open(os.path.join(testpath, 'inlist'), 'w') as test_inlist:
                for one_line in default_inlist.readlines():
                    if ("max_model_number" in one_line):
                        test_inlist.write("         max_model_number = "+str(number_of_steps+1)+"\n")
                    elif ("mesa_data_dir" in one_line):
                        test_inlist.write("      mesa_data_dir = '"+os.path.join(mesa_src_path, 'data')+"'\n")
                    elif ("zams_filename" in one_line):
                        test_inlist.write("      zams_filename = '"+os.path.join(instance.get_output_directory(), "star_data", "starting_models", "zams_z20m3.data")+"'\n")
                    else:
                        test_inlist.write(one_line)
        instance.stop()

        (stdout, stderr) = Popen([mesa_star_path], cwd=testpath, stdin=PIPE, stdout=PIPE, stderr=PIPE).communicate()
        self.assertEqual(stderr, b"")
        for i, line in enumerate(stdout.decode("utf-8").splitlines()):
            # print i, line, line in amuse_output
            if i == 52 + 4 * number_of_steps + 8 * (number_of_steps//10) + (3 if number_of_steps > 55 else 0):
                self.assertEqual(line, "stop because model_number >= max_model_number")
                break
            else:
                self.assertTrue(line in amuse_output)

    def test18(self):
        print("Testing MESA mass_change (User-specified wind/accretion)")
        instance = self.new_instance_of_an_optional_code(MESA)
        instance.parameters.RGB_wind_scheme = 0  # must be turned off for user-specified rates
        instance.parameters.AGB_wind_scheme = 0  # must be turned off for user-specified rates

        star = instance.particles.add_particle(Particle(mass=1 | units.MSun))
        star.mass_change = 1.0e-8 | units.MSun / units.yr  # positive -> accretion
        star.evolve_one_step()

        self.assertAlmostRelativeEqual(star.mass_change, 1.0e-8 | units.MSun / units.yr)
        self.assertAlmostRelativeEqual(star.wind, -1.0e-8 | units.MSun / units.yr, 3)
        self.assertAlmostRelativeEqual(star.age, 1.0e5 | units.yr)
        self.assertAlmostRelativeEqual(star.mass, 1.0010 | units.MSun)

        star.mass_change = -1.0e-8 | units.MSun / units.yr  # negative -> wind
        star.evolve_one_step()

        self.assertAlmostRelativeEqual(star.mass_change, -1.0e-8 | units.MSun / units.yr)
        self.assertAlmostRelativeEqual(star.wind, 1.0e-8 | units.MSun / units.yr, 3)
        self.assertAlmostRelativeEqual(star.age, 1.8e5 | units.yr)
        self.assertAlmostRelativeEqual(star.mass, 1.0002 | units.MSun)
        print(star.as_set())
        instance.stop()

    def slowtest19a(self):
        print("Testing MESA core mass")
        instance = self.new_instance_of_an_optional_code(MESA)
        star = instance.particles.add_particle(Particle(mass=3 | units.MSun))
        star.evolve_for(330 | units.Myr)
        for i in range(10):
            star.evolve_for(10 | units.Myr)
            index = numpy.searchsorted(star.get_chemical_abundance_profiles(number_of_species=1)[0], 1.0e-4)
            h_poor_mass = EnclosedMassInterpolator(radii=star.get_radius_profile(), densities=star.get_density_profile()).enclosed_mass[index].as_quantity_in(units.MSun)
            print(h_poor_mass, star.core_mass)
            self.assertAlmostEqual(star.core_mass, h_poor_mass, 2)
        instance.stop()

    def test19b(self):
        print("Testing MESA core mass (short version of slowtest19a)")
        instance = self.new_instance_of_an_optional_code(MESA)
        star = instance.particles.add_particle(Particle(mass=3 | units.MSun))
        star.evolve_one_step()
        index = numpy.searchsorted(star.get_chemical_abundance_profiles(number_of_species=1)[0], 1.0e-4)
        h_poor_mass = EnclosedMassInterpolator(radii=star.get_radius_profile(), densities=star.get_density_profile()).enclosed_mass[index].as_quantity_in(units.MSun)
        self.assertAlmostEqual(star.core_mass, h_poor_mass, 2)
        instance.stop()

    def test20(self):
        print("Testing MESA pre-main-sequence star")
        instance = self.new_instance_of_an_optional_code(MESA)
        star = instance.pre_ms_stars.add_particle(Particle(mass=1.0 | units.MSun))

        self.assertAlmostEqual(star.time_step, 1.0e-3 | units.yr)
        star.evolve_one_step()
        self.assertAlmostEqual(star.age, 1.0e-3 | units.yr)

        instance.evolve_model(1.0 | units.yr)
        self.assertEqual(star.stellar_type, 17 | units.stellar_type)
        self.assertEqual(str(star.stellar_type), "Pre-main-sequence Star")
        self.assertTrue(star.age > 1.0 | units.yr)
        self.assertTrue(star.temperature < 4500 | units.K)
        self.assertTrue(star.luminosity > 10 | units.LSun)
        self.assertTrue(star.radius > 2 | units.RSun)
        instance.stop()

    def slowtest21(self):
        print("Testing MESA calculate_core_mass")
        instance = self.new_instance_of_an_optional_code(MESA)
        star = instance.particles.add_particle(Particle(mass=40 | units.MSun))
        instance.evolve_model(4.56 | units.Myr)
        total_core_mass = star.calculate_core_mass()
        self.assertAlmostRelativeEqual(star.core_mass, total_core_mass, 2)
        self.assertTrue(star.calculate_core_mass(core_H_abundance_limit=1.0e-2) > total_core_mass)
        self.assertEqual(star.calculate_core_mass(core_H_abundance_limit=1.0e-4), total_core_mass)
        self.assertTrue(star.calculate_core_mass(core_H_abundance_limit=1.0e-6) < total_core_mass)
        self.assertAlmostRelativeEqual(star.calculate_core_mass(core_H_abundance_limit=0.8), star.mass, 2)

        h1_core_mass = star.calculate_core_mass(species=["h1"])
        he3_core_mass = star.calculate_core_mass(species=["he3"])
        he4_core_mass = star.calculate_core_mass(species=["he4"])
        c12_core_mass = star.calculate_core_mass(species=["c12"])
        n14_core_mass = star.calculate_core_mass(species=["n14"])
        o16_core_mass = star.calculate_core_mass(species=["o16"])
        ne20_core_mass = star.calculate_core_mass(species=["ne20"])
        mg24_core_mass = star.calculate_core_mass(species=["mg24"])
        metal_core_mass = star.calculate_core_mass(species=["c12", "n14", "o16", "ne20", "mg24"])
        print(h1_core_mass, he3_core_mass, he4_core_mass, c12_core_mass, n14_core_mass, o16_core_mass)
        print(ne20_core_mass, mg24_core_mass, metal_core_mass)
        self.assertAlmostRelativeEqual(star.core_mass, total_core_mass, 2)
        instance.stop()
        self.assertAlmostRelativeEqual(total_core_mass, he4_core_mass, 1)
        self.assertAlmostRelativeEqual(total_core_mass, he4_core_mass + metal_core_mass, 4)
        self.assertAlmostRelativeEqual(total_core_mass, he4_core_mass + metal_core_mass + h1_core_mass, 7)
        self.assertAlmostRelativeEqual(metal_core_mass,
            c12_core_mass + n14_core_mass + o16_core_mass + ne20_core_mass + mg24_core_mass, 7)
        self.assertAlmostEqual(he3_core_mass, 0 | units.MSun)

    def test22(self):
        print("Testing MESA calculate_core_mass (short version of slowtest21)")
        instance = self.new_instance_of_an_optional_code(MESA)
        star = instance.particles.add_particle(Particle(mass=1 | units.MSun))
        instance.evolve_model(0.3 | units.Gyr)  # VERY short, for test speed up
        central_hydrogen_abundance = star.get_chemical_abundance_profiles()[0][0]
        self.assertTrue(central_hydrogen_abundance < 0.68)  # some hydrogen is burned
        self.assertTrue(central_hydrogen_abundance > 0.67)  # ... but not that much yet
        self.assertEqual(star.calculate_core_mass(core_H_abundance_limit=0.67), 0 | units.MSun)
        self.assertAlmostEqual(star.calculate_core_mass(core_H_abundance_limit=0.70), 1 | units.MSun, 3)

        # For test speed up, we use a weird core_H_abundance_limit to define the "hydrogen exhausted core"
        limit = 0.68
        expected_core_mass = 0.01786033709 | units.MSun
        self.assertAlmostEqual(star.calculate_core_mass(core_H_abundance_limit=limit), expected_core_mass, 3)

        h1_core_mass = star.calculate_core_mass(species=["h1"], core_H_abundance_limit=limit)
        he3_core_mass = star.calculate_core_mass(species=["he3"], core_H_abundance_limit=limit)
        he4_core_mass = star.calculate_core_mass(species=["he4"], core_H_abundance_limit=limit)
        c12_core_mass = star.calculate_core_mass(species=["c12"], core_H_abundance_limit=limit)
        n14_core_mass = star.calculate_core_mass(species=["n14"], core_H_abundance_limit=limit)
        o16_core_mass = star.calculate_core_mass(species=["o16"], core_H_abundance_limit=limit)
        ne20_core_mass = star.calculate_core_mass(species=["ne20"], core_H_abundance_limit=limit)
        mg24_core_mass = star.calculate_core_mass(species=["mg24"], core_H_abundance_limit=limit)
        metal_core_mass = star.calculate_core_mass(species=["c12", "n14", "o16", "ne20", "mg24"], core_H_abundance_limit=limit)
        instance.stop()
        self.assertAlmostRelativeEqual(h1_core_mass, expected_core_mass*0.68, 2)
        self.assertAlmostRelativeEqual(he4_core_mass, expected_core_mass*0.30, 2)
        self.assertAlmostRelativeEqual(metal_core_mass, expected_core_mass*0.02, 1)
        self.assertAlmostRelativeEqual(expected_core_mass, he4_core_mass + he3_core_mass + metal_core_mass + h1_core_mass, 7)
        self.assertAlmostRelativeEqual(metal_core_mass,
            c12_core_mass + n14_core_mass + o16_core_mass + ne20_core_mass + mg24_core_mass, 7)
        self.assertAlmostEqual(he3_core_mass, 0 | units.MSun, 5)

    def test23(self):
        print("Testing MESA central_temperature and central_density")
        instance = self.new_instance_of_an_optional_code(MESA)
        stars = instance.particles.add_particles(Particles(mass=[0.1, 1, 10] | units.MSun))
        self.assertIsOfOrder(stars.central_temperature, [4e6, 13e6, 31e6] | units.K)
        self.assertIsOfOrder(stars.central_density, [400, 77, 9] | units.g * units.cm**-3)
        instance.stop()

    def test24(self):
        print("Testing MESA calculate_helium_exhausted_core_mass")
        instance = self.new_instance_of_an_optional_code(MESA)
        star = instance.particles.add_particle(Particle(mass=2 | units.MSun))

        composition = star.get_chemical_abundance_profiles()
        # Mimic hydrogen exhausted core:
        composition[2, :100] = composition[2, :100] + composition[0, :100]
        composition[0, :100] = 0
        # Mimic helium exhausted core:
        carbon_oxygen = composition[:6, :50].sum(axis=0)
        composition[3, :50] = carbon_oxygen * 0.6
        composition[5, :50] = carbon_oxygen * 0.4
        composition[1:3, :50] = 0
        composition[4, :50] = 0
        star.set_chemical_abundance_profiles(composition)

        self.assertAlmostRelativeEqual(star.calculate_core_mass(),
            star.mass * star.get_cumulative_mass_profile()[100], 1)
        self.assertAlmostRelativeEqual(star.calculate_helium_exhausted_core_mass(),
            star.mass * star.get_cumulative_mass_profile()[50], 1)

        core_mass = star.calculate_helium_exhausted_core_mass(split_species=False)
        core_mass_by_species = star.calculate_helium_exhausted_core_mass(split_species=True)
        carbon_mass_in_core, oxygen_mass_in_core = star.calculate_helium_exhausted_core_mass(
            split_species=True, species=["c12", "o16"])

        self.assertEqual(len(core_mass_by_species), len(star.get_names_of_species()))
        instance.stop()
        self.assertAlmostRelativeEqual(core_mass, core_mass_by_species.sum())
        self.assertEqual(core_mass_by_species[0:3].sum(), 0 | units.MSun)
        self.assertEqual(core_mass_by_species[4], 0 | units.MSun)
        self.assertEqual(core_mass_by_species[3], carbon_mass_in_core)
        self.assertEqual(core_mass_by_species[5], oxygen_mass_in_core)

    def test25(self):
        print("Testing MESA accretion")
        instance = self.new_instance_of_an_optional_code(MESA)
        instance.parameters.RGB_wind_scheme = 0
        instance.parameters.AGB_wind_scheme = 0
        star = instance.particles.add_particle(Particle(mass=2 | units.MSun))

        self.assertEqual(star.get_accrete_same_as_surface(), 1)
        star.set_accrete_same_as_surface(0)
        self.assertEqual(star.get_accrete_same_as_surface(), 0)

        self.assertEqual(star.get_accrete_composition_non_metals(), [-1.0, -1.0, -1.0, -1.0])
        self.assertEqual(star.get_accrete_composition_metals_identifier(), -1)
        self.assertEqual(star.get_accrete_composition_metals(), [-1.0]*28)
        print("Accreting 75% deuterium", end=' ')
        composition_light = [0, 0.75, 0, 0]
        print("and 25% iron")
        composition_metals = [0]*23 + [1.0] + [0]*4
        star.set_accrete_composition_non_metals(*composition_light)
        star.set_accrete_composition_metals_identifier(0)  # i.e. specified below:
        star.set_accrete_composition_metals(*composition_metals)
        self.assertEqual(star.get_accrete_composition_non_metals(), composition_light)
        self.assertEqual(star.get_accrete_composition_metals_identifier(), 0)
        self.assertEqual(star.get_accrete_composition_metals(), composition_metals)

        star.mass_change = 1.0e-8 | units.MSun / units.yr
        star.time_step = 0.1 | units.yr
        instance.evolve_model(1 | units.yr)
        composition = star.get_chemical_abundance_profiles()
        species = star.get_names_of_species()
        print("Both deuterium and iron are not in the current net,")
        print("so have been added to {0} and {1}".format(species[0], species[-1]))
        self.assertEqual(composition[:, -1], [0.75, 0, 0, 0, 0, 0, 0, 0.25])
        instance.stop()



from amuse.test import amusetest

from amuse.support.exceptions import AmuseException
from amuse import datamodel
from amuse.units import units
from amuse.units import nbody_system

import numpy


class TestParticleLinkToParticle(amusetest.TestCase):
    """
    Tests One-to-One relation between particles
    """

    def test1(self):

        particles = datamodel.Particles(3)
        particles.mass = [2, 3, 4] | units.kg

        parent = particles[0]
        child1 = particles[1]
        child2 = particles[2]

        parent.child1 = child1
        parent.child2 = child2
        child1.sibling = child2

        self.assertEqual(len(particles.child1), 3)
        self.assertEqual(len(particles.child1.as_set().compressed()), 1)
        self.assertEqual(len(particles.child2), 3)
        self.assertEqual(len(particles.child2.as_set().compressed()), 1)

        self.assertAlmostRelativeEqual(parent.child1.mass, 3 | units.kg)

    def test2(self):

        particles = datamodel.Particles(3)
        particles.mass = [2, 3, 4] | units.kg

        parent = particles[0]
        child1 = particles[1]
        child2 = particles[2]

        parent.child1 = child1
        parent.child2 = child2
        child1.sibling = child2

        convert_nbody = nbody_system.nbody_to_si(10 | units.kg, 5 | units.m)
        particles = datamodel.ParticlesWithUnitsConverted(
            particles,
            convert_nbody.as_converter_from_generic_to_si()
        )
        self.assertEqual(len(particles.child1), 3)
        self.assertEqual(len(particles.child1.as_set().compressed()), 1)
        self.assertEqual(len(particles.child2), 3)
        self.assertEqual(len(particles.child2.as_set().compressed()), 1)

        self.assertAlmostRelativeEqual(particles[0].child1.mass, 0.3 | nbody_system.mass)
        self.assertAlmostRelativeEqual(particles[0].child1.mass, 0.3 | nbody_system.mass)

    def test3(self):

        particles = datamodel.Particles(3)
        particles.mass = [2, 3, 4] | units.kg

        parent = particles[0]
        child1 = particles[1]
        child2 = particles[2]

        parent.child1 = child1
        parent.child2 = child2
        child1.sibling = child2

        other = datamodel.Particles()
        particles.synchronize_to(other)
        root = None
        for x in other:
            if not x.child1 is None:
                root = x
                break

        self.assertFalse(root is None)

        self.assertAlmostRelativeEqual(root.child1.mass, 3 | units.kg)
        self.assertTrue(root.child1.as_set()._original_set() is other)
        self.assertTrue(root.child1.sibling.as_set()._original_set() is other)

    def test4(self):

        particles = datamodel.Particles(3)
        particles.mass = [2, 3, 4] | units.kg
        other = particles.copy()

        parent = particles[0]
        child1 = particles[1]
        child2 = particles[2]

        parent.child1 = child1
        parent.child2 = child2
        child1.sibling = child2

        channel = particles.new_channel_to(other)
        channel.copy()

        self.assertTrue(other[0].child1.as_set()._original_set() is other)
        self.assertTrue(other[1].sibling.as_set()._original_set() is other)
        self.assertAlmostRelativeEqual(other[0].child1.mass, 3 | units.kg)

    def test5(self):

        particles = datamodel.Particles(3)
        parent = particles[0]
        child1 = particles[1]
        child2 = particles[2]

        particles.mass = [2, 3, 4] | units.kg

        parent.child1 = child1
        parent.child2 = child2
        child1.sibling = child2

        other = particles.copy()

        self.assertAlmostRelativeEqual(other[0].child1.mass, 3 | units.kg)
        self.assertTrue(other[0].child1.as_set()._original_set() is other)
        self.assertTrue(other[1].sibling.as_set()._original_set() is other)

    def test6(self):

        binaries = datamodel.Particles(2)
        binaries.mass = [4, 6] | units.kg

        stars = datamodel.Particles(4)
        stars.mass = [1, 3, 4, 2] | units.kg

        binaries[0].child1 = stars[0]
        binaries[0].child2 = stars[1]
        binaries[1].child1 = stars[3]
        binaries[1].child2 = stars[2]

        other = binaries.copy()
        stars.mass = 2 * ([1, 3, 4, 2] | units.kg)
        self.assertAlmostRelativeEqual(binaries[0].child1.mass, 2 | units.kg)
        self.assertAlmostRelativeEqual(other[0].child1.mass, 1 | units.kg)
        self.assertAlmostRelativeEqual(other[0].child2.mass, 3 | units.kg)
        self.assertAlmostRelativeEqual(binaries[1].child1.mass, 4 | units.kg)
        self.assertAlmostRelativeEqual(other[1].child1.mass, 2 | units.kg)
        self.assertAlmostRelativeEqual(other[1].child2.mass, 4 | units.kg)

    def test7(self):

        binaries = datamodel.Particles(2)
        binaries.mass = [4, 6] | units.kg

        stars = datamodel.Particles(4)
        stars.mass = [1, 3, 4, 2] | units.kg

        binaries[0].child1 = stars[0]
        binaries[0].child2 = stars[1]
        binaries[1].child1 = stars[3]
        binaries[1].child2 = stars[2]
        stars[0].binary = binaries[0]
        stars[1].binary = binaries[0]
        stars[2].binary = binaries[1]
        stars[3].binary = binaries[1]

        other = binaries.copy()
        stars.mass *= 2
        binaries.mass *= 2

        self.assertAlmostRelativeEqual(other[0].child1.mass, 1 | units.kg)
        self.assertAlmostRelativeEqual(binaries[0].child1.mass, 2 | units.kg)
        self.assertAlmostRelativeEqual(other[0].child2.mass, 3 | units.kg)
        self.assertAlmostRelativeEqual(other[1].child1.mass, 2 | units.kg)
        self.assertAlmostRelativeEqual(other[1].child2.mass, 4 | units.kg)
        self.assertAlmostRelativeEqual(binaries[0].child1.binary.mass, 8 | units.kg)
        self.assertAlmostRelativeEqual(other[0].child1.binary.mass, 4 | units.kg)
        self.assertAlmostRelativeEqual(other[1].child1.binary.mass, 6 | units.kg)

    def test8(self):

        particles = datamodel.Particles(2)
        particles.mass = [2, 3] | units.kg

        binary = datamodel.Particles(1)
        binary[0].mass = 5 | units.kg
        binary[0].child1 = particles[0]
        binary[0].child2 = particles[1]

        binaries = datamodel.Particles()
        binaries.add_particles(binary)

        self.assertAlmostRelativeEqual(binaries[0].child1.mass, 2 | units.kg)
        self.assertAlmostRelativeEqual(binaries[0].child2.mass, 3 | units.kg)

    def test9(self):
        particles1 = datamodel.Particles(keys=[9, 10, 11, 12])
        particles1.mass = [1, 2, 3, 4] | units.kg

        particles2 = datamodel.Particles(keys=[13, 14, 15])
        particles2.mass = [20, 21, 22] | units.kg
        particles1[0].particle2 = particles2[0]
        particles1[1].particle2 = particles2[0]

        particles1_copy = particles1.copy()

        self.assertEqual(particles1_copy[0].particle2.key,  13)
        self.assertEqual(particles1_copy[1].particle2.key,  13)

        particles1_copy[0].particle2.mass = 30 | units.kg

        self.assertAlmostRelativeEquals(particles1_copy[0].particle2.mass, 30 | units.kg)
        self.assertAlmostRelativeEquals(particles1_copy[1].particle2.mass, 30 | units.kg)


class TestParticleLinkToParticles(amusetest.TestCase):
    """
    Tests One-to-Many relation between a particle and particle-sets
    """

    def test0(self):
        particles1 = datamodel.Particles(keys=[9, 10, 11, 12])
        particles1.mass = [1, 2, 3, 4] | units.kg
        x = numpy.empty((10,), dtype=object)
        index = numpy.asarray([1])
        x[index] = particles1
        print(x)
        self.assertEqual(len(x[1]), 4)

    def test1(self):
        particles1 = datamodel.Particles(keys=[9, 10, 11, 12])
        particles1.mass = [1, 2, 3, 4] | units.kg
        particles2 = datamodel.Particles(keys=[13, 14])
        particles2.mass = [20, 21] | units.kg
        particles1[1].particles2 = particles2
        self.assertEqual(particles1[1].particles2.key,  [13, 14])
        self.assertAlmostRelativeEquals(particles1[1].particles2.mass, [20, 21] | units.kg)
        self.assertEqual(particles1[0].particles2, None)

    def test2(self):
        particles1 = datamodel.Particles(keys=[9, 10, 11, 12])
        particles1.mass = [1, 2, 3, 4] | units.kg
        particles2 = datamodel.Particles(keys=[13, 14])
        particles2.mass = [20, 21] | units.kg
        particles1[1].particles2 = particles2
        self.assertEqual(len(particles1.particles2), 4)
        self.assertEqual(particles1.particles2[1].key,  [13, 14])
        self.assertEqual(particles1.particles2[0],  None)

    def test3(self):
        particles1 = datamodel.Particles(keys=[9, 10, 11, 12])
        particles1.mass = [1, 2, 3, 4] | units.kg
        particles2 = datamodel.Particles(keys=[13, 14])
        particles2.mass = [20, 21] | units.kg
        particles1[1].particles2 = particles2

        particles1_copy = particles1.copy()
        self.assertEqual(particles1_copy[1].particles2.key,  [13, 14])
        self.assertAlmostRelativeEquals(particles1_copy[1].particles2.mass, [20, 21] | units.kg)
        self.assertEqual(particles1_copy[0].particles2, None)
        self.assertNotEqual(id(particles1_copy[1].particles2), id(particles1[1].particles2))

    def test4(self):
        particles1 = datamodel.Particles(keys=[9, 10, 11, 12])
        particles1.mass = [1, 2, 3, 4] | units.kg
        particles1_copy = particles1.copy()

        particles2 = datamodel.Particles(keys=[13, 14])
        particles2.mass = [20, 21] | units.kg
        particles1[1].particles2 = particles2

        channel = particles1.new_channel_to(particles1_copy)
        channel.copy()

        self.assertEqual(particles1_copy[1].particles2.key,  [13, 14])
        self.assertAlmostRelativeEquals(particles1_copy[1].particles2.mass, [20, 21] | units.kg)
        self.assertEqual(particles1_copy[0].particles2, None)
        self.assertEqual(id(particles1_copy[1].particles2), id(particles1[1].particles2))

    def test5(self):
        particles1 = datamodel.Particles(keys=[9, 10, 11, 12])
        particles1.mass = [1, 2, 3, 4] | units.kg

        particles2 = datamodel.Particles(keys=[13, 14])
        particles2.mass = [20, 21] | units.kg
        particles1.particles2 = [None, particles2, None, particles2[0]]

        self.assertEqual(particles1[1].particles2.key,  [13, 14])
        self.assertEqual(particles1[1].particles2.mass, [20, 21] | units.kg)
        self.assertAlmostRelativeEquals(particles1[3].particles2.key,  13)
        self.assertAlmostRelativeEquals(particles1[3].particles2.mass, 20 | units.kg)
        self.assertEqual(particles1[0].particles2, None)
        self.assertEqual(particles1[2].particles2, None)

    def test6(self):
        particles1 = datamodel.Particles(keys=[9, 10, 11, 12])
        particles1.mass = [1, 2, 3, 4] | units.kg

        particles2 = datamodel.Particles(keys=[13, 14, 15])
        particles2.mass = [20, 21, 22] | units.kg
        particles1[0].particles2 = particles2[0:2]
        particles1[1].particles2 = particles2[1:]

        particles1_copy = particles1.copy(keep_structure=False)

        self.assertEqual(particles1_copy[0].particles2.key,  [13, 14])
        self.assertEqual(particles1_copy[1].particles2.key,  [14, 15])

        particles1_copy[0].particles2[1].mass = 30 | units.kg

        self.assertAlmostRelativeEquals(particles1_copy[0].particles2[1].mass, 30 | units.kg)
        self.assertAlmostRelativeEquals(particles1_copy[1].particles2[0].mass, 21 | units.kg)

        particles1_copy = particles1.copy(keep_structure=True)

        self.assertEqual(particles1_copy[0].particles2.key,  [13, 14])
        self.assertEqual(particles1_copy[1].particles2.key,  [14, 15])

        particles1_copy[0].particles2[1].mass = 30 | units.kg

        self.assertAlmostRelativeEquals(particles1_copy[0].particles2[1].mass, 30 | units.kg)
        self.assertAlmostRelativeEquals(particles1_copy[1].particles2[0].mass, 30 | units.kg)


class TestParticleLinkToGridPoint(amusetest.TestCase):
    """
    Tests One-to-One relation between particles and gridpoints
    """

    def test1(self):

        particles = datamodel.Particles(3)
        particles.mass = [2, 3, 4] | units.kg
        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3

        particles[0].point11 = grid[1][1]

        self.assertAlmostRelativeEquals(grid[1][1].rho, 6 | units.kg / units.m**3)
        self.assertAlmostRelativeEquals(particles[0].point11.rho, 6 | units.kg / units.m**3)

    def test2(self):

        particles = datamodel.Particles(3)
        particles.mass = [2, 3, 4] | units.kg
        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3
        particles[0].point11 = grid[1][1]

        particles_copy = particles.copy()

        self.assertAlmostRelativeEquals(grid[1][1].rho, 6 | units.kg / units.m**3)
        self.assertAlmostRelativeEquals(particles_copy[0].point11.rho, 6 | units.kg / units.m**3)

    def test3(self):

        particles = datamodel.Particles(3)
        particles.mass = [2, 3, 4] | units.kg
        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3

        particles_copy = particles.copy()

        particles[0].point11 = grid[1][1]

        channel = particles.new_channel_to(particles_copy)
        channel.copy()

        self.assertAlmostRelativeEquals(grid[1][1].rho, 6 | units.kg / units.m**3)
        self.assertAlmostRelativeEquals(particles_copy[0].point11.rho, 6 | units.kg / units.m**3)


class TestParticleLinkToGrids(amusetest.TestCase):
    """
    Tests One-to-Many relation between particles and grids
    """

    def test1(self):

        particles = datamodel.Particles(3)
        particles.mass = [2, 3, 4] | units.kg
        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3

        particles[0].grid = grid

        self.assertAlmostRelativeEquals(grid[1][1].rho, 6 | units.kg / units.m**3)
        self.assertAlmostRelativeEquals(particles[0].grid[1][1].rho, 6 | units.kg / units.m**3)

    def test2(self):

        particles = datamodel.Particles(3)
        particles.mass = [2, 3, 4] | units.kg
        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3

        particles[0].grid = grid

        self.assertAlmostRelativeEquals(grid[1][1].rho, 6 | units.kg / units.m**3)
        self.assertAlmostRelativeEquals(particles[0].grid[1][1].rho, 6 | units.kg / units.m**3)

        particles_copy = particles.copy()

        grid[1][1].rho = 10 | units.kg / units.m**3

        self.assertAlmostRelativeEquals(grid[1][1].rho, 10 | units.kg / units.m**3)
        self.assertAlmostRelativeEquals(particles[0].grid[1][1].rho, 10 | units.kg / units.m**3)
        self.assertAlmostRelativeEquals(particles_copy[0].grid[1][1].rho, 6 | units.kg / units.m**3)

    def test3(self):

        particles = datamodel.Particles(3)
        particles.mass = [2, 3, 4] | units.kg
        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3

        particles_copy = particles.copy()

        particles[0].grid = grid

        self.assertAlmostRelativeEquals(grid[1][1].rho, 6 | units.kg / units.m**3)
        self.assertAlmostRelativeEquals(particles[0].grid[1][1].rho, 6 | units.kg / units.m**3)

        channel = particles.new_channel_to(particles_copy)
        channel.copy()

        self.assertAlmostRelativeEquals(particles_copy[0].grid[1][1].rho, 6 | units.kg / units.m**3)


class TestGridPointLinkToParticle(amusetest.TestCase):
    """
    Tests One-to-One relation between gridpoints and particles
    """

    def test1(self):

        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3
        particles = datamodel.Particles(3)
        particles.mass = [2, 3, 4] | units.kg

        grid[0][0].particle = particles[1]

        self.assertAlmostRelativeEquals(grid[0][0].particle.mass, 3 | units.kg)
        self.assertEqual(grid[0][0].particle, particles[1])
        self.assertEqual(grid[1][1].particle, None)

    def test2(self):

        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3
        particles = datamodel.Particles(3)
        particles.mass = [2, 3, 4] | units.kg

        grid[0][0].particle = particles[1]

        grid_copy = grid.copy()

        self.assertAlmostRelativeEquals(grid_copy[0][0].particle.mass, 3 | units.kg)
        self.assertEqual(grid_copy[0][0].particle, particles[1])
        self.assertEqual(grid_copy[1][1].particle, None)

        grid[0][0].particle.mass = 5 | units.kg

        self.assertAlmostRelativeEquals(grid_copy[0][0].particle.mass, 3 | units.kg)

    def test3(self):

        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3
        particles = datamodel.Particles(3)
        particles.mass = [2, 3, 4] | units.kg

        grid_copy = grid.copy()

        grid[0][0].particle = particles[1]

        channel = grid.new_channel_to(grid_copy)
        channel.copy()

        self.assertAlmostRelativeEquals(grid_copy[0][0].particle.mass, 3 | units.kg)
        self.assertEqual(grid_copy[0][0].particle, particles[1])
        self.assertEqual(grid_copy[1][1].particle, None)


class TestGridPointLinkToParticles(amusetest.TestCase):
    """
    Tests One-to-Many relation between gridpoints and particles
    """

    def test1(self):

        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3
        particles = datamodel.Particles(3)
        particles.mass = [2, 3, 4] | units.kg

        grid[0][0].particles = particles

        self.assertAlmostRelativeEquals(grid[0][0].particles[1].mass, 3 | units.kg)
        self.assertEqual(grid[0][0].particles[1], particles[1])
        self.assertEqual(grid[1][1].particles, None)

    def test2(self):

        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3
        particles = datamodel.Particles(3)
        particles.mass = [2, 3, 4] | units.kg

        grid[0][0].particles = particles

        grid_copy = grid.copy()

        self.assertAlmostRelativeEquals(grid_copy[0][0].particles[1].mass, 3 | units.kg)

        grid[0][0].particles[1].mass = 10 | units.kg

    def test3(self):

        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3
        particles = datamodel.Particles(3)
        particles.mass = [2, 3, 4] | units.kg

        grid_copy = grid.copy()
        grid[0][0].particles = particles

        channel = grid.new_channel_to(grid_copy)
        channel.copy()

        self.assertAlmostRelativeEquals(grid_copy[0][0].particles[1].mass, 3 | units.kg)

        grid[0][0].particles[1].mass = 10 | units.kg

        self.assertAlmostRelativeEquals(grid_copy[0][0].particles[1].mass, 10 | units.kg)


class TestGridPointLinkToGridPoint(amusetest.TestCase):
    """
    Tests One-to-One relation between gridpoints
    """

    def test1(self):

        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3

        grid[0][0].neighbour = grid[0][1]

        self.assertAlmostRelativeEquals(grid[0][0].neighbour.rho, 3 | units.kg / units.m**3)
        self.assertEqual(grid[0][0].neighbour, grid[0][1])
        self.assertEqual(grid[1][1].neighbour, None)

    def test2(self):

        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3

        grid[0][0].neighbour = grid[0][1]

        grid_copy = grid.copy()

        self.assertAlmostRelativeEquals(grid_copy[0][0].neighbour.rho, 3 | units.kg / units.m**3)
        self.assertEqual(grid_copy[0][0].neighbour, grid_copy[0][1])
        self.assertEqual(grid_copy[1][1].neighbour, None)

        grid[0][0].neighbour.rho = 5 | units.kg / units.m**3

        self.assertAlmostRelativeEquals(grid_copy[0][0].neighbour.rho, 3 | units.kg / units.m**3)

        grid_copy[0][1].rho = 6 | units.kg / units.m**3

        self.assertAlmostRelativeEquals(grid_copy[0][0].neighbour.rho, 6 | units.kg / units.m**3)

    def test3(self):

        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3

        grid_copy = grid.copy()

        grid[0][0].neighbour = grid[0][1]

        channel = grid.new_channel_to(grid_copy)
        channel.copy()

        self.assertAlmostRelativeEquals(grid_copy[0][0].neighbour.rho, 3 | units.kg / units.m**3)
        self.assertEqual(grid_copy[0][0].neighbour, grid_copy[0][1])
        self.assertEqual(grid_copy[1][1].neighbour, None)

        grid[0][0].neighbour.rho = 5 | units.kg / units.m**3

        self.assertAlmostRelativeEquals(grid_copy[0][0].neighbour.rho, 3 | units.kg / units.m**3)

        grid_copy[0][1].rho = 6 | units.kg / units.m**3

        self.assertAlmostRelativeEquals(grid_copy[0][0].neighbour.rho, 6 | units.kg / units.m**3)


class TestGridPointLinkToGrid(amusetest.TestCase):
    """
    Tests One-to-Many relation between gridpoints
    """

    def test1(self):

        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3

        grid[0][0].container = grid

        self.assertAlmostRelativeEquals(grid[0][0].container[0][1].rho, 3 | units.kg / units.m**3)
        self.assertEqual(grid[0][0].container[0][1], grid[0][1])
        self.assertEqual(grid[1][1].container, None)

    def test2(self):

        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3

        grid[0][0].container = grid

        grid_copy = grid.copy()

        self.assertAlmostRelativeEquals(grid_copy[0][0].container[0][1].rho, 3 | units.kg / units.m**3)
        self.assertEqual(grid_copy[0][0].container[0][1], grid_copy[0][1])
        self.assertEqual(grid_copy[1][1].container, None)

        grid[0][0].container[0][1].rho = 5 | units.kg / units.m**3

        self.assertAlmostRelativeEquals(grid_copy[0][0].container[0][1].rho, 3 | units.kg / units.m**3)

        grid_copy[0][1].rho = 6 | units.kg / units.m**3

        self.assertAlmostRelativeEquals(grid_copy[0][0].container[0][1].rho, 6 | units.kg / units.m**3)

    def test3(self):

        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3

        grid_copy = grid.copy()

        grid[0][0].container = grid

        channel = grid.new_channel_to(grid_copy)
        channel.copy()

        self.assertAlmostRelativeEquals(grid_copy[0][0].container[0][1].rho, 3 | units.kg / units.m**3)
        self.assertEqual(grid_copy[0][0].container[0][1], grid_copy[0][1])
        self.assertEqual(grid_copy[1][1].container, None)

        grid[0][0].container[0][1].rho = 5 | units.kg / units.m**3

        self.assertAlmostRelativeEquals(grid_copy[0][0].container[0][1].rho, 3 | units.kg / units.m**3)

        grid_copy[0][1].rho = 6 | units.kg / units.m**3

        self.assertAlmostRelativeEquals(grid_copy[0][0].container[0][1].rho, 6 | units.kg / units.m**3)

    def test4(self):

        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3

        grid[..., 0].container = grid

        self.assertEqual(id(grid[0][0].container), id(grid))
        self.assertEqual(id(grid[1][0].container), id(grid))
        self.assertEqual(grid[0][1].container, None)
        self.assertEqual(grid[1][1].container, None)

    def test5(self):

        grid = datamodel.Grid(2, 3)
        grid.rho = [[2, 3, 4], [5, 6, 7]] | units.kg / units.m**3

        grid.container = grid

        for index in numpy.ndindex(*grid.shape):
            self.assertEqual(id(grid[index].container), id(grid))
            self.assertEqual(grid[index].rho, (index[0] * 3 + index[1] + 2) | units.kg / units.m**3)



from amuse.test import amusetest
import numpy

from amuse.support.exceptions import AmuseException


from amuse.units import core
from amuse.units.units import *
from amuse.units.constants import *


class TestUnitConversions(amusetest.TestCase):
    def test1(self):
        km = 1000 * m
        self.assertEqual(1000, km.value_in(m))
        self.assertEqual(0.001, m.value_in(km))

    def test2(self):
        km = 1000 * m
        val = km(10)
        self.assertEqual(10000, val.value_in(m))

    def test3(self):
        km = 1000.0 * m
        val = km(10)
        self.assertEqual("10 1000.0 * m", str(val))

        km = named('kilometer', 'km', 1000 * m)
        val = km(10)
        self.assertEqual("10 km", str(val))

    def test4(self):
        km = named('kilometer', 'km', 1000 * m)
        h = named('hour', 'h', 60 * 60 * s)
        kmh = km/h
        ms = m/s
        val = 10 | m/s
        self.assertEqual(36, val.value_in(kmh))

    def test5(self):
        km = named('kilometer', 'km', 1000 * m)
        h = named('hour', 'h', 60 * 60 * s)
        kmh = km/h
        ms = m/s
        val = 10 | ms
        self.assertEqual(36, val.value_in(kmh))

    def test6(self):
        no1 = m / m
        no2 = no_unit
        self.assertEqual("m / m", str(no1))
        self.assertTrue(no1.has_same_base_as(no2))

    def test7(self):
        x = (100 * kg ** 2) / kg
        self.assertEqual("100 * kg**2 / kg", str(x))
        self.assertEqual("100.0 * kg", str(x.to_simple_form()))

    def test8(self):
        x = (10 | g) * (2 | m)
        self.assertEqual("20 0.001 * m * kg", str(x))
        self.assertEqual("0.02 kg * m", str(x.as_quantity_in(kg * m)))
        x = (10 | kg) * (2000 | g)
        self.assertEqual("20000 0.001 * kg**2", str(x))

    def test9(self):
        speed_of_light = 1 | (lightyear * yr**-1)
        time = 1e-9 | s
        length = speed_of_light * time
        length_in_m = length.value_in(m)
        self.assertAlmostEqual(0.2997988, length_in_m, 6)

    def test10(self):
        eps0_1 = mu0**-1*c**-2
        eps0_2 = (1 | none)/(mu0*(c**2))
        self.assertTrue((eps0_1 - eps0_2) / eps0_1 < (1e-10 | none))
        b = ((1. | e)**2)
        f = (hbar*c*4.*numpy.pi * eps0)**-1
        fine_structure_constant_calculated = (b * f)
        fine_structure_constant = 7.297352537650e-3
        self.assertAlmostEqual(fine_structure_constant_calculated, fine_structure_constant, 5)

    def test11(self):
        vel1 = 1 | m / s
        self.assertRaises(core.IncompatibleUnitsException, vel1.as_quantity_in, s / m,
            expected_message="Cannot express m / s in s / m, the units do not have the same bases")

    def test12(self):
        self.assertEqual((1234 | g).as_string_in(g), '1234 g')
        self.assertEqual((1234 | g).as_string_in(kg), '1.234 kg')
        self.assertEqual((1.234 | kg).as_string_in(g), '1234.0 g')
        self.assertEqual((1.0 | km * s**-1).as_string_in(m / s), '1000.0 m / s')
        self.assertEqual((1.0 | km * s**-1).as_string_in(s**-1 * m), '1000.0 s**-1 * m')
        self.assertEqual((1.0 | km / s).as_string_in((10*J/g)**0.5), '10.0 (10 * J / g)**0.5')

    def test13(self):
        self.assertAlmostRelativeEquals((1234 | km ** (5/3.)).value_in((km ** 5.0) ** (1.0/3.0)), 1234.0, 12)
        self.assertAlmostRelativeEquals((1234 | km ** (5/3.)).value_in((m ** 5.0) ** (1.0/3.0)), 123400000.0)

    def test15(self):
        self.assertAlmostRelativeEquals((m / s).to_array_of_floats(), [1, 1,  1, 0, -1, 0, 0, 0, 0])
        self.assertAlmostRelativeEquals((s / m).to_array_of_floats(), [1, 1, -1, 0, 1, 0, 0, 0, 0])
        self.assertAlmostRelativeEquals((kg ** 2 / s).to_array_of_floats(), [1, 1,  0, 2, -1, 0, 0, 0, 0])


class TestNonNumericUnits(amusetest.TestCase):
    def test1(self):
        string1 = "string" | string
        self.assertRaises(AmuseException, string1.as_quantity_in, m,
            expected_message="Cannot convert non-numeric quantities in to another unit")

    def test2(self):
        x = "test" | string
        self.assertEqual("test", x.value_in(string))

    def test3(self):
        test_unit = core.enumeration_unit(
            "test",
            "test",
            [1, 2, 3],
            ["one", "two", "three"]
        )
        x = 1 | test_unit
        self.assertEqual(1, x.value_in(test_unit))
        self.assertEqual("one", str(x))
        self.assertRaises(Exception, lambda: 4 | test_unit,
            expected_message="<4> is not a valid value for unit<test>")

    def test4(self):
        self.assertRaises(Exception, lambda: 1 | string,
            expected_message="<1> is not a valid value for unit<string>")

    def test5(self):
        test_unit = core.enumeration_unit(
            "test",
            "test",
            [1, 2, 3],
            ["one", "two", "three"]
        )
        self.assertEqual(3, len(list(test_unit.quantities())))
        for x, y in zip(test_unit.quantities(), ["one", "two", "three"]):
            self.assertEqual(str(x), y)

    def test6(self):
        test_unit = core.enumeration_unit(
            "test",
            "test",
            [1, 4, 7]
        )
        self.assertEqual(3, len(list(test_unit.quantities())))
        for x, y in zip(test_unit.quantities(), ["1", "4", "7"]):
            self.assertEqual(str(x), y)

    def test7(self):
        test_unit = core.enumeration_unit(
            "test",
            "test",
            range(5)
        )
        self.assertEqual(5, len(list(test_unit.quantities())))
        self.assertEqual(1 | test_unit, 1 | test_unit)
        self.assertTrue(1 | test_unit == 1 | test_unit)
        self.assertFalse(1 | test_unit == 2 | test_unit)
        self.assertTrue(1 | test_unit != 2 | test_unit)
        self.assertFalse(1 | test_unit != 1 | test_unit)
        self.assertTrue(1 | test_unit >= 1 | test_unit)
        self.assertFalse(1 | test_unit >= 2 | test_unit)
        self.assertTrue(1 | test_unit <= 1 | test_unit)
        self.assertFalse(1 | test_unit <= 0 | test_unit)
        self.assertTrue(1 | test_unit > 0 | test_unit)
        self.assertFalse(1 | test_unit > 1 | test_unit)
        self.assertTrue(1 | test_unit < 3 | test_unit)
        self.assertFalse(1 | test_unit < 0 | test_unit)



#!/usr/bin/env python3
from support.classifiers import classifiers

from setuptools import setup

import support
support.use("system")
from support.setup_codes import setup_commands

name = 'amuse-tests'
author = 'The AMUSE team'
author_email = 'info@amusecode.org'
license_ = "Apache License 2.0"
url = 'http://www.amusecode.org/'
install_requires = [
    'amuse-framework',
]
description = 'The Astrophysical Multipurpose Software Environment - tests'
with open("README.md", "r") as fh:
    long_description = fh.read()
long_description_content_type = "text/markdown"

extensions = []

all_data_files = []

packages = [
    'amuse.test.suite',
    'amuse.test.suite.ext_tests', 
    'amuse.test.suite.core_tests',
    'amuse.test.suite.compile_tests', 
    'amuse.test.suite.codes_tests',
    'amuse.test.suite.ticket_tests', 
    'amuse.test.suite.reports',
]

package_data = {
}

mapping_from_command_name_to_command_class = setup_commands()

setup_requires = ['setuptools_scm']
use_scm_version = {
    "root": "../..",
    "relative_to": __file__,
    "version_file": "src/amuse/test/suite/_version.py",
}

setup(
    name=name,
    use_scm_version=use_scm_version,
    setup_requires=setup_requires,
    classifiers=classifiers,
    url=url,
    author_email=author_email,
    author=author,
    license=license_,
    description=description,
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    install_requires=install_requires,
    python_requires=">=3.7",
    cmdclass=mapping_from_command_name_to_command_class,
    ext_modules=extensions,
    package_dir={
        'amuse.test.suite': 'src/amuse/test/suite',
    },
    packages=packages,
    package_data=package_data,
    data_files=all_data_files,
)
